from copy import deepcopy
from typing import List
from rdkit import Chem
from icolos.core.step_utils.obabel_structconvert import OBabelStructConvert

from icolos.utils.enums.compound_enums import (
    CompoundContainerEnum,
    EnumerationContainerEnum,
)
from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum
from icolos.core.step_utils.structconvert import StructConvert
from icolos.utils.general.icolos_exceptions import ContainerCorrupted
from icolos.utils.enums.write_out_enums import WriteOutEnum
from typing import Union
import numpy as np
import os

_WE = WriteOutEnum()


class Conformer:
    """This class is a storage class for individual conformers associated with a given Enumeration."""

    def __init__(
        self,
        conformer: Chem.Mol = None,
        conformer_id: int = None,
        enumeration_object=None,
    ):
        self._conformer = conformer
        self._conformer_id = conformer_id
        self._enumeration_object = enumeration_object
        self._extra_data_dictionary = {}

    def get_compound_name(self) -> str:
        if self.get_enumeration_object() is not None:
            return self.get_enumeration_object().get_compound_name()

    def get_index_string(self) -> str:
        enum_obj = self.get_enumeration_object()
        enum_str = ""
        if enum_obj is not None:
            enum_str = enum_obj.get_index_string()
        conf_str = ""
        if self.get_conformer_id() is not None:
            conf_str = str(self.get_conformer_id())
        return ":".join([enum_str, conf_str])

    def add_extra_data(self, key: str, data):
        self._extra_data_dictionary[key] = data

    def get_extra_data(self) -> dict:
        return self._extra_data_dictionary

    def clear_extra_data(self):
        self._extra_data_dictionary = {}

    def set_enumeration_object(self, enumeration_object):
        self._enumeration_object = enumeration_object

    def get_enumeration_object(self):
        return self._enumeration_object

    def get_molecule(self) -> Chem.Mol:
        return self._conformer

    def set_molecule(self, conformer: Chem.Mol):
        self._conformer = conformer

    def set_conformer_id(self, conformer_id: int):
        self._conformer_id = conformer_id

    def get_conformer_id(self) -> int:
        return self._conformer_id

    def empty(self) -> bool:
        if self.get_molecule() is None:
            return True
        return False

    def _clone(self):
        clone = Conformer(
            conformer=deepcopy(self.get_molecule()),
            conformer_id=self.get_conformer_id(),
            enumeration_object=self.get_enumeration_object(),
        )
        clone._extra_data_dictionary = deepcopy(self.get_extra_data())
        return clone

    def __copy__(self):
        return self._clone()

    def __deepcopy__(self, memo):
        return self._clone()

    def __repr__(self):
        parent_enumeration_id = (
            None
            if self.get_enumeration_object() is None
            else self.get_enumeration_object().get_enumeration_id()
        )
        return "<Icolos conformer: id=%s, parent enumeration: %s>" % (
            self.get_conformer_id(),
            parent_enumeration_id,
        )

    def __str__(self):
        return self.__repr__()

    def write(self, path: str, format_=_WE.SDF):
        writer = Chem.SDWriter(path)
        molecule = self.get_molecule()
        molecule.SetProp(_WE.RDKIT_NAME, self.get_index_string())
        molecule.SetProp(_WE.INDEX_STRING, self.get_index_string())
        writer.write(molecule)
        writer.close()
        if format_ == _WE.PDB:
            pdb_path = path.split(".")[0] + ".pdb"
            # convert the written sdf file to a pdb with OB
            converter = OBabelStructConvert()
            converter.sdf2pdb(sdf_file=path, pdb_file=pdb_path)
            os.remove(path)

    def update_coordinates(self, path: str):
        old = self.get_molecule()
        for mol in Chem.SDMolSupplier(path, removeHs=False):
            mol.SetProp(_WE.RDKIT_NAME, old.GetProp(_WE.RDKIT_NAME))
            for prop in old.GetPropNames():
                mol.SetProp(prop, old.GetProp(prop))
            self.set_molecule(mol)

            # only one molecule expected at this stage, so stop after first run
            break
        self.write("".join([path, "_out"]))


class Enumeration:
    """This class bundles all information on an enumeration, especially all conformers generated."""

    def __init__(
        self,
        compound_object=None,
        smile: str = "",
        molecule: Chem.Mol = None,
        original_smile: str = None,
        enumeration_id: int = None,
    ):
        self._MC = CompoundContainerEnum()
        self._EC = EnumerationContainerEnum()
        self._smile = smile
        self._compound_object = compound_object
        self._molecule = molecule
        self._original_smile = original_smile
        self._enumeration_id = enumeration_id
        self._conformers = []

    def empty(self) -> bool:
        if len(self.get_conformers()) == 0:
            return True
        return False

    def get_compound_name(self) -> str:
        if self.get_compound_object() is not None:
            return self.get_compound_object().get_name()

    def _get_next_conformer_id(self) -> int:
        ids = [conf.get_conformer_id() for conf in self.get_conformers()]
        if len(ids) == 0:
            return 0
        else:
            return max(ids) + 1

    def sort_conformers(
        self, by_tag: Union[str, List[str]], reverse: bool = True, aggregation="sum"
    ):
        conformers = self.get_conformers()
        if isinstance(by_tag, str):
            conformers = sorted(
                conformers,
                key=lambda x: float(x.get_molecule().GetProp(by_tag)),
                reverse=reverse,
            )
            self._conformers = conformers
            self.reset_conformer_ids()
        elif isinstance(by_tag, list):
            # need to normalise the values, calculate max and min of each tag in the series
            def normalise_tag(value, tag):
                all_tag_values = [
                    float(conf.get_molecule().GetProp(tag)) for conf in conformers
                ]
                max_tag = np.max(all_tag_values)
                min_tag = np.min(all_tag_values)
                return (float(value) - min_tag) / (max_tag - min_tag)

            # if we specify multiple tags, aggregate according the the provided aggregation function
            if aggregation == "sum":
                conformers = sorted(
                    conformers,
                    key=lambda x: np.sum(
                        [
                            float(normalise_tag(x.get_molecule().GetProp(i), i))
                            for i in by_tag
                        ]
                    ),
                    reverse=reverse,
                )
                self._conformers = conformers
            elif aggregation == "product":
                conformers = sorted(
                    conformers,
                    key=lambda x: np.product(
                        [
                            float(normalise_tag(x.get_molecule().GetProp(i), i))
                            for i in by_tag
                        ]
                    ),
                    reverse=reverse,
                )
                self._conformers = conformers
            else:
                raise AttributeError(
                    "Only sum or product aggregation modes are currently supported - ABORT"
                )
                # for ligand in self.ligands:

    #    ligand.set_conformers(sorted(ligand.get_conformers(),
    #                                 key=lambda x: float(x.GetProp(_ROE.GLIDE_DOCKING_SCORE)), reverse=False))
    #    ligand.add_tags_to_conformers()

    def find_conformer(self, conformer_id: int) -> Conformer:
        conf = [
            conf
            for conf in self.get_conformers()
            if conf.get_conformer_id() == conformer_id
        ]
        if len(conf) == 0:
            raise IndexError(f"Could not find conformer with id {conformer_id}.")
        elif len(conf) > 1:
            raise ContainerCorrupted(
                f"More than one conformer with id {conformer_id} found in the same Enumeration instance (compound_number: {self.get_enumeration_id()})."
            )
        return conf[0]

    def get_conformer_ids(self) -> List[int]:
        ids = [conf.get_conformer_id() for conf in self.get_conformers()]
        return ids

    def reset_conformer_ids(self):
        for new_id, conf in enumerate(self.get_conformers()):
            conf.set_conformer_id(conformer_id=new_id)

    def add_conformer(self, conformer: Conformer, auto_update: bool = True):
        """Add a new conformer. If "auto_update" is True, the Enumeration class will be set to "self" and
        the conformer_id will be set to the next free index."""
        conformer = deepcopy(conformer)
        if auto_update:
            conformer.set_enumeration_object(self)
            conformer.set_conformer_id(self._get_next_conformer_id())
        self._conformers.append(conformer)

    def add_conformers(self, conformers: List[Conformer], auto_update: bool = True):
        """Add new conformers. If "auto_update" is True, the Enumeration class will be set to "self" and
        the conformer_id will be set to the next free index."""
        for conformer in conformers:
            self.add_conformer(conformer=conformer, auto_update=auto_update)

    def get_index_string(self) -> str:
        comp_obj = self.get_compound_object()
        comp_str = ""
        if comp_obj is not None:
            comp_str = comp_obj.get_index_string()
        enum_str = ""
        if self.get_enumeration_id() is not None:
            enum_str = str(self.get_enumeration_id())
        return ":".join([comp_str, enum_str])

    def clean_failed_conformers(self):
        # all conformers, where the molecule has been set to None by a function can be considered to have failed
        for idx in list(reversed(range(len(self._conformers)))):
            if self._conformers[idx].get_molecule() is None:
                del self._conformers[idx]
        self.reset_conformer_ids()

    def clear_molecule(self):
        self._molecule = None

    def clear_conformers(self):
        self._conformers = []

    def get_conformers(self) -> List[Conformer]:
        return self._conformers

    def clone_conformers(self) -> List[Conformer]:
        return [deepcopy(conf) for conf in self._conformers]

    def set_compound_object(self, compound_object):
        self._compound_object = compound_object

    def get_compound_object(self):
        return self._compound_object

    def set_enumeration_id(self, enumeration_id: int):
        self._enumeration_id = enumeration_id

    def get_enumeration_id(self) -> int:
        return self._enumeration_id

    def set_smile(self, smile: str):
        self._smile = smile

    def get_smile(self) -> str:
        return self._smile

    def set_molecule(self, molecule: Chem.Mol):
        self._molecule = molecule

    def get_molecule(self) -> Chem.Mol:
        return self._molecule

    def set_original_smile(self, original_smile: str):
        self._original_smile = original_smile

    def get_original_smile(self) -> str:
        return self._original_smile

    def _clone(self):
        clone = Enumeration(
            compound_object=self.get_compound_object(),
            smile=self.get_smile(),
            molecule=deepcopy(self.get_molecule()),
            original_smile=self.get_original_smile(),
            enumeration_id=self.get_enumeration_id(),
        )
        for conf in self.get_conformers():
            conf = deepcopy(conf)
            conf.set_enumeration_object(enumeration_object=clone)
            clone.add_conformer(conf, auto_update=False)
        return clone

    def __copy__(self):
        return self._clone()

    def __deepcopy__(self, memo):
        return self._clone()

    def __repr__(self):
        parent_compound_id = (
            None
            if self.get_compound_object() is None
            else self.get_compound_object().get_compound_number()
        )
        return (
            "<Icolos enumeration: id=%s, smile=%s, parent compound: %s, num_conformers: %i>"
            % (
                self.get_enumeration_id(),
                self.get_smile(),
                parent_compound_id,
                len(self._conformers),
            )
        )

    def __str__(self):
        return self.__repr__()

    def __iter__(self):
        return iter(self._conformers)

    def __getitem__(self, key: int) -> Conformer:
        return self._conformers[key]

    def __len__(self) -> int:
        return len(self.get_conformers())


class Compound:
    """This class bundles all information on a molecule and serves mainly to group enumerations."""

    def __init__(self, name: str = "", compound_number: int = None):
        self._CC = CompoundContainerEnum()
        self._EC = EnumerationContainerEnum()
        self._name = name
        self._compound_number = compound_number
        self._enumerations = []

    def __repr__(self):
        return "<Icolos compound: name=%s, compound_number=%s, enumerations=%s>" % (
            self.get_name(),
            self.get_compound_number(),
            len(self.get_enumerations()),
        )

    def __str__(self):
        return self.__repr__()

    def get_index_string(self) -> str:
        if self.get_compound_number() is not None:
            return str(self.get_compound_number())
        else:
            return ""

    def set_name(self, name: str):
        self._name = name

    def get_name(self) -> str:
        return self._name

    def set_compound_number(self, compound_number: int):
        self._compound_number = compound_number

    def get_compound_number(self) -> int:
        return self._compound_number

    def add_enumeration(self, enumeration: Enumeration, auto_update: bool = True):
        """Add a new enumeration. If "auto_update" is True, the Compound class will be set to "self" and
        the enumeration_id will be set to the next free index."""
        enumeration = deepcopy(enumeration)
        if auto_update:
            enumeration.set_compound_object(self)
            enumeration.set_enumeration_id(self._get_next_enumeration_id())
        self._enumerations.append(enumeration)

    def add_enumerations(
        self, enumerations: List[Enumeration], auto_update: bool = True
    ):
        """Add new enumerations. If "auto_update" is True, the Compound class will be set to "self" and
        the enumeration_id will be set to the next free index."""
        for enumeration in enumerations:
            self.add_enumeration(enumeration=enumeration, auto_update=auto_update)

    def clear_enumerations(self):
        self._enumerations = []

    def find_enumeration(self, idx: int):
        for enum in self.get_enumerations():
            if enum.get_enumeration_id() == idx:
                return enum

    def get_enumerations(self) -> List[Enumeration]:
        return self._enumerations

    def _clone(self):
        clone = Compound(
            name=self.get_name(), compound_number=self.get_compound_number()
        )
        for enum in self.get_enumerations():
            enum = deepcopy(enum)
            enum.set_compound_object(compound_object=clone)
            clone.add_enumeration(enum, auto_update=False)
        return clone

    def __iter__(self):
        return iter(self._enumerations)

    def __copy__(self):
        return self._clone()

    def __deepcopy__(self, memo):
        return self._clone()

    def __getitem__(self, key: int) -> Enumeration:
        return self._enumerations[key]

    def __len__(self) -> int:
        return len(self.get_enumerations())

    def _get_next_enumeration_id(self):
        ids = [enum.get_enumeration_id() for enum in self.get_enumerations()]
        if len(ids) == 0:
            return 0
        else:
            return max(ids) + 1

    def find_enumeration(self, enumeration_id: int) -> Enumeration:
        enum = [
            enum
            for enum in self.get_enumerations()
            if enum.get_enumeration_id() == enumeration_id
        ]
        if len(enum) == 0:
            raise IndexError(f"Could not find enumeration with id {enumeration_id}.")
        elif len(enum) > 1:
            raise ContainerCorrupted(
                f"More than one enumeration with id {enumeration_id} found in the same Compound instance (compound_number: {self.get_compound_number()})."
            )
        return enum[0]

    def get_enumeration_ids(self) -> List[int]:
        ids = [enum.get_enumeration_id() for enum in self.get_enumerations()]
        return ids

    def reset_enumeration_ids(self):
        for new_id, enum in enumerate(self.get_enumerations()):
            enum.set_enumeration_id(enumeration_id=new_id)

    def reset_all_ids(self):
        self.reset_enumeration_ids()
        for enum in self.get_enumerations():
            enum.reset_conformer_ids()

    def update_all_relations(self):
        for enum in self.get_enumerations():
            enum.set_compound_object(self)
            for conf in enum.get_conformers():
                conf.set_enumeration_object(enum)

    def empty(self) -> bool:
        if len(self.get_enumerations()) == 0:
            return True
        return False

    def unroll_conformers(self) -> List[Conformer]:
        conformers = []
        for enum in self.get_enumerations():
            # guard against empty enumerations that might be used when constructing more complex data flows
            if enum.empty():
                continue
            for conf in enum.get_conformers():
                conformers.append(conf)
        return conformers


# TODO: Replacing these three functions by a wrapper object
def get_compound_by_id(compounds: List[Compound], id: int) -> Compound:
    for compound in compounds:
        if compound.get_compound_number() == id:
            return compound
    raise ValueError(
        f"Could not find compound with id {id} in list of length {len(compounds)}."
    )


def get_compound_by_name(compounds: List[Compound], name: str) -> Compound:
    for compound in compounds:
        if compound.get_name() == name:
            return compound
    raise ValueError(
        f"Could not find compound with name {name} in list of length {len(compounds)}."
    )


def unroll_conformers(compounds: List[Compound]) -> List[Conformer]:
    all_conformers = []
    for comp in compounds:
        all_conformers = all_conformers + comp.unroll_conformers()
    return all_conformers


def unroll_enumerations(compounds: List[Compound]) -> List[Enumeration]:
    all_enumerations = []
    for comp in compounds:
        all_enumerations = all_enumerations + comp.get_enumerations()
    return all_enumerations



import logging

from icolos.loggers.base_logger import BaseLogger


class BlankLogger(BaseLogger):
    """This logger serves as a "verbatim" interface."""

    def __init__(self):
        super().__init__()

    def _initialize_logger(self):
        logger = logging.getLogger(self._LE.LOGGER_BLANK)
        return logger



import json
import os
import re
from copy import deepcopy
from tempfile import mkdtemp
from typing import List, Tuple
from pydantic import BaseModel

from icolos.core.containers.compound import Conformer
from icolos.core.workflow_steps.calculation.base import StepCalculationBase
from icolos.utils.enums.program_parameters import JazzyEnum
from icolos.utils.enums.step_enums import StepJazzyEnum
from icolos.core.workflow_steps.step import _LE
from icolos.utils.execute_external.jazzy import JazzyExecutor
from icolos.utils.general.files_paths import gen_tmp_file
from icolos.utils.general.parallelization import SubtaskContainer, Parallelizer

_SJE = StepJazzyEnum()
_JE = JazzyEnum()

_all_jazzy_commands = [_JE.VEC, _JE.VIS]

_all_jazzy_properties = [
    _JE.RESULT_DGA,
    _JE.RESULT_DGP,
    _JE.RESULT_DGTOT,
    _JE.RESULT_SA,
    _JE.RESULT_SDC,
    _JE.RESULT_SDX,
]


class JazzyAdditional(BaseModel):
    command: str = _JE.VEC


class StepJazzy(StepCalculationBase, BaseModel):

    jazzy_additional: JazzyAdditional = None

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=JazzyExecutor)
        self._check_backend_availability()

        # initialize the additional settings
        self.jazzy_additional = JazzyAdditional(**self.settings.additional)
        if self.jazzy_additional.command not in _all_jazzy_commands:
            raise ValueError(
                f"Jazzy command {self.jazzy_additional.command} unknown - abort."
            )

    def _prepare_batch(self, batch) -> Tuple:
        tmp_dirs = []
        input_smiles = []
        output_files = []
        conformers = []
        for next_subtask_list in batch:
            tmp_dir = mkdtemp()
            tmp_dirs.append(tmp_dir)
            for subtask in next_subtask_list:
                _, tmp_out_path = gen_tmp_file(suffix=".out", dir=tmp_dir)
                output_files.append(tmp_out_path)
                conformer = subtask.data
                conformers.append(conformer)

                # TODO: not really efficient for SMILES, but keep it (?) on conformer level in anticipation of
                #       structural input that differs per conformer
                input_smile = conformer.get_enumeration_object().get_smile()
                input_smiles.append(input_smile)
        return tmp_dirs, input_smiles, output_files, conformers

    def _prepare_arguments(self, settings: List) -> List:

        # add flags
        for flag in self.settings.arguments.flags:
            settings.append(flag)

        # add parameters
        parameters = deepcopy(self.settings.arguments.parameters)

        # flatten the dictionary into a list for command-line execution
        for key in parameters.keys():
            if key in _all_jazzy_commands:
                self._logger.log(
                    f"Use the additional block to specify Jazzy commands, parameter {key} ignored.",
                    _LE.WARNING,
                )
                continue
            settings.append(key)
            settings.append(parameters[key])
        return settings

    def _run_subjob(self, tmp_dir: str, input_smile: str, output_file: str) -> None:
        work_dir = os.getcwd()
        os.chdir(tmp_dir)

        # construct the specified command-line call; only one command can be used at a time
        # e.g. jazzy vec [OPTIONS] SMILES
        arguments = [self.jazzy_additional.command, '"' + input_smile + '"']
        arguments = self._prepare_arguments(arguments)

        result = self._backend_executor.execute(
            command=_JE.JAZZY, arguments=arguments, check=False
        )

        # Jazzy prints the result to stdout -> store it in a temporary file
        with open(output_file, "w") as f:
            f.writelines(result.stdout)

        os.chdir(work_dir)

    def _parse_jazzy_result(
        self, output_files: List[str], conformers: List[Conformer]
    ) -> List:

        results = []

        for output_file, conformer in zip(output_files, conformers):
            # load the JSON string that was captured from stdout and written to the output file
            try:
                with open(output_file) as file:
                    # Jazzy does not output valid JSONs (' instead of "), so we need to replace those
                    # except escaped ones
                    result = file.read().replace("\r", "").replace("\n", "")
                    p = re.compile("(?<!\\\\)'")
                    result = p.sub('"', result)
                    result = json.loads(result)
            except FileNotFoundError:
                self._logger.log(
                    f"Jazzy result for conformer {conformer.get_index_string()} stored in file {output_file} not found - proceeding.",
                    _LE.WARNING,
                )
                results.append(_SJE.FAILURE)
                continue

            # attach the properties obtained as tags
            for key in result.keys():
                if key in _all_jazzy_properties:
                    conformer.get_molecule().SetProp(key, str(result[key]))
            results.append(_SJE.SUCCESS)
        return results

    def _execute_jazzy(self):
        jazzy_parallelizer = Parallelizer(func=self._run_subjob)
        n = 1
        while self._subtask_container.done() is False:
            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())
            tmp_dirs, input_smiles, output_files, conformers = self._prepare_batch(
                next_batch
            )

            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]

            self._logger.log(f"Executing Jazzy for batch {n}.", _LE.DEBUG)

            jazzy_parallelizer.execute_parallel(
                tmp_dir=tmp_dirs, input_smile=input_smiles, output_file=output_files
            )

            results = self._parse_jazzy_result(output_files, conformers)
            for sublist, result in zip(next_batch, results):
                assert len(sublist) == 1
                for task in sublist:
                    if result == _SJE.SUCCESS:
                        task.set_status_success()
                    else:
                        task.set_status_failed()
            n += 1
            self._remove_temporary(tmp_dirs)

    def execute(self):
        all_conformers = []
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if enumeration.get_conformers():
                    for conformer in enumeration.get_conformers():
                        all_conformers.append(conformer)
        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(all_conformers)
        self._execute_jazzy()
        self._logger.log(
            f"Completed execution of Jazzy for {len(all_conformers)} conformers (using their SMILES strings).",
            _LE.INFO,
        )



#!/usr/bin/env python
#  coding=utf-8

import os
import sys
import json
import argparse
from datetime import datetime
from icolos.core.composite_agents.workflow import WorkFlow


from icolos.utils.enums.composite_agents_enums import WorkflowEnum
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.entry_points import ExecutorEnum

from icolos.utils.entry_point_functions.logging_helper_functions import (
    initialize_logging,
)
from icolos.utils.entry_point_functions.parsing_functions import (
    parse_header,
    log_version_number,
    get_version_number,
    version_match,
    get_versions_as_strings,
)
from icolos.utils.general.citation_generator import print_citations
from icolos.utils.general.files_paths import attach_root_path


def main():
    # enums
    _LE = LoggingConfigEnum()
    _WE = WorkflowEnum()

    # get the input parameters and parse them
    version = get_version_number()
    parser = argparse.ArgumentParser(
        description='Implements entry point for "Icolos" workflow execution.',
        epilog=f"Icolos version: {'unknown' if version is None else version}",
    )
    parser.add_argument(
        "-conf",
        type=str,
        default=None,
        help="A path to an workflow's configuration file (JSON dictionary) that is to be executed.",
    )
    parser.add_argument(
        "-debug",
        action="store_true",
        help="Set this flag to activate the inbuilt debug logging mode.",
    )
    parser.add_argument(
        "--global_variables",
        nargs="+",
        default=None,
        type=str,
        help='List of strings, setting global variables with key and value, e.g. "root:/path/to/root".',
    )
    parser.add_argument(
        "--global_settings",
        nargs="+",
        default=None,
        type=str,
        help='List of strings, setting global settings with key and value, e.g. "remove_temporary_files:False" "single_directory:True".',
    )
    args, args_unk = parser.parse_known_args()

    if args.conf is None or not os.path.isfile(args.conf):
        raise Exception(
            'Parameter "-conf" must be a relative or absolute path to a configuration (JSON) file.'
        )

    # load configuration
    with open(args.conf) as file:
        conf = file.read().replace("\r", "").replace("\n", "")
        conf = json.loads(conf)

    # set the logging configuration according to parameters
    log_conf = attach_root_path(_LE.PATH_CONFIG_DEFAULT)
    if args.debug:
        log_conf = attach_root_path(_LE.PATH_CONFIG_DEBUG)
    logger = initialize_logging(log_conf_path=log_conf, workflow_conf=conf)

    # write the version of the installation used to the logfile and check match with configuration
    log_version_number(logger)
    if not version_match(conf):
        version_config, version_installation = get_versions_as_strings(conf)
        logger.log(
            f"Version of configuration file ({version_config}) and installation ({version_installation}) do not match or are not defined, proceed with caution.",
            _LE.WARNING,
        )

    # update global variables and settings
    conf = parse_header(
        conf=conf, args=args, entry_point_path=os.path.realpath(__file__), logger=logger
    )
    # generate workflow object
    workflow = WorkFlow(**conf[_WE.WORKFLOW])
    workflow.initialize()

    # print the logo and ,depending on the backend software used, the citations for the workflow given
    print_citations(workflow._initialized_steps)

    # execute the whole workflow
    st_time = datetime.now()
    workflow.execute()
    exec_time = datetime.now() - st_time
    logger.log(f"Icolos workflow completed. Walltime: {exec_time}.", _LE.INFO)


if __name__ == "__main__":
    main()

    sys.exit(0)



from typing import List
from pydantic.main import BaseModel
from icolos.core.workflow_steps.step import StepBase
from icolos.utils.general.parallelization import Parallelizer, SubtaskContainer
from icolos.core.workflow_steps.step import _LE


class IterParallelizer(BaseModel):
    # config block controlling how the steps are parallelized
    # if you are executing a 5 step workflow with 10 repeats, dependent_steps = 5, cores = 10
    # this will allow each independent replica to be allocated to a single job queue, retaining step order
    parallelize: bool = False
    jobs: int = 1
    dependent_steps: int = None


class StepDispatcher(StepBase, BaseModel):
    """
    Step class containing job control functionality required for StepIterator, supports Slurm for job scheduling
    Supports running Icolos process as master job for parallel step execution on cluster.  Generates a pool of initialized steps to be executed, based on the

    Step-type class for disaptching multiple steps in parallel, useful for executing multiple batch jobs simultaneously
    """

    workflows: List = []
    # expect the parallel execution block to be handed over from flow control
    parallel_execution: IterParallelizer = IterParallelizer()

    def __init__(self, **data):
        super().__init__(**data)

    def _prepare_batch(self, batch) -> List[List[StepBase]]:

        batch_steps = []
        for sublist in batch:
            sublist_steps = []
            for task in sublist:
                sublist_steps.append(task.data)
            batch_steps.append(sublist_steps)
        return batch_steps

    def execute(self):
        """
        Execute multiple steps in parallel
        """
        # Spin up multiple processes.
        self.execution.parallelization.jobs = self.parallel_execution.jobs

        # TODO, we can repeat entire workflows if we want, I'm not sure this makes sense though
        self._subtask_container = SubtaskContainer(max_tries=1)
        self._subtask_container.load_data(self.workflows)

        parallelizer = Parallelizer(func=self.execute_workflow)
        n = 1

        while self._subtask_container.done() is False:

            next_batch = self._get_sublists(
                get_first_n_lists=self.parallel_execution.jobs
            )  # return n lists of length max_sublist_length
            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]

            self._logger.log(
                f"Starting {len(next_batch)} parallel jobs under Icolos JobControl, execution batch {n}",
                _LE.INFO,
            )

            jobs = self._prepare_batch(next_batch)

            parallelizer.execute_parallel(jobs=jobs)

            # TODO: sucessful execution of each step is not explicitly checked,
            # the step is responsible for throwing errors if something has gone wrong
            for task in next_batch:
                for subtask in task:
                    subtask.set_status_success()

    def execute_workflow(self, jobs):
        # submits then monitors the step
        wf_data = self.get_workflow_object().workflow_data
        for idx, job in enumerate(jobs):
            # copy existing wf data up to this point int othe new wf object
            job.initialize()
            job.workflow_data = wf_data
            self._logger.log(f"Executing workflow {idx} of {len(jobs)}", _LE.DEBUG)
            job.execute()



import os
from collections import OrderedDict
from copy import deepcopy

import numpy as np
import pandas as pd
import json
from typing import List
from pydantic import BaseModel, PrivateAttr
from rdkit import Chem
from pathlib import Path

# from icolos.core.composite_agents.workflow import WorkflowData

from icolos.core.containers.compound import Compound, Conformer
from icolos.core.step_utils.input_preparator import StepData
from icolos.core.step_utils.run_variables_resolver import RunVariablesResolver
from icolos.loggers.steplogger import StepLogger
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.step_enums import StepBaseEnum, StepGromacsEnum
from icolos.utils.enums.write_out_enums import WriteOutEnum

_WE = WriteOutEnum()
_LE = LoggingConfigEnum()
_SBE = StepBaseEnum
_SGE = StepGromacsEnum()


class StepWriteoutCompoundAggregationParameters(BaseModel):
    mode: _SBE = _SBE.WRITEOUT_COMP_AGGREGATION_MODE_ALL
    highest_is_best: bool = True
    key: str = None


class StepWriteoutCompoundParameters(BaseModel):
    category: _SBE
    aggregation: StepWriteoutCompoundAggregationParameters = (
        StepWriteoutCompoundAggregationParameters()
    )
    key: str = None
    selected_tags: List[str] = None


class StepWriteoutGenericParameters(BaseModel):
    key: str


class StepWriteoutGromacsParameters(BaseModel):
    key: str


class StepWriteoutDestinationParameters(BaseModel):
    resource: str = None
    type: _SBE = _SBE.WRITEOUT_DESTINATION_TYPE_FILE
    format: _SBE = _SBE.FORMAT_TXT
    merge: bool = True
    mode: _SBE = _SBE.WRITEOUT_DESTINATION_BASE_NAME


class StepWriteoutParameters(BaseModel):
    compounds: StepWriteoutCompoundParameters = None
    generic: StepWriteoutGenericParameters = None
    gmx_state: StepWriteoutGromacsParameters = None
    destination: StepWriteoutDestinationParameters = None


class WriteOutHandler(BaseModel):

    config: StepWriteoutParameters
    data: StepData = None
    workflow_data: BaseModel = None

    class Config:
        underscore_attrs_are_private = True

    _logger = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._logger = StepLogger()

    def set_data(self, data: StepData):
        self.data = deepcopy(data)

    def set_workflow_data(self, data):
        self.workflow_data = data

    def get_data(self) -> StepData:
        return self.data

    def _handle_destination_type(self):
        if self.config.destination.type.lower() in (
            _SBE.WRITEOUT_DESTINATION_TYPE_FILE,
            _SBE.WRITEOUT_DESTINATION_TYPE_REINVENT,
            _SBE.WRITEOUT_DESTINATION_DIR,
        ):
            return self.config.destination.resource
        elif (
            self.config.destination.type.lower() == _SBE.WRITEOUT_DESTINATION_TYPE_REST
        ):
            raise ValueError("REST end-point destination type not supported yet.")
        raise ValueError(
            f"Destination type {self.config.destination.type} not supported."
        )

    def _write_compounds(self):
        resource = self._handle_destination_type()
        resolver = RunVariablesResolver()
        if self.config.compounds.category == _SBE.WRITEOUT_COMP_CATEGORY_CONFORMERS:
            if self.config.destination.format.upper() == _SBE.FORMAT_CSV:
                if self.config.destination.format.upper() != _SBE.FORMAT_CSV:
                    raise NotImplementedError(
                        "Only supporting CSV write-out format for tabular data."
                    )
                self._writeout_tabular()
            elif self.config.destination.format.upper() == _SBE.FORMAT_JSON:
                self._writeout_reinvent()
            elif self.config.destination.format.upper() == _SBE.FORMAT_SDF:

                def _write_compounds(compounds: List[Compound], resource: str):
                    # TODO: deal with resolving resources differently (also for writing enumerations below)
                    resource_resolved = resource
                    for compound in compounds:
                        for enum in compound.get_enumerations():
                            if len(enum.get_conformers()) > 0:
                                resource_resolved = resolver.resolve(resource, enum[0])
                                break
                    self._make_folder(resource_resolved)

                    writer = Chem.SDWriter(resource_resolved)
                    written = 0

                    for comp in compounds:
                        for enum in comp:
                            for conf in enum:
                                molecule = conf.get_molecule()
                                if (
                                    comp.get_name() is not None
                                    and comp.get_name() != ""
                                ):
                                    molecule.SetProp(_WE.COMPOUND_NAME, comp.get_name())
                                molecule.SetProp(
                                    _WE.RDKIT_NAME, conf.get_index_string()
                                )
                                molecule.SetProp(
                                    _WE.INDEX_STRING, conf.get_index_string()
                                )
                                writer.write(molecule)
                                written += 1
                    writer.close()
                    self._logger.log(
                        f"Wrote {written} conformers to file {resource_resolved}.",
                        _LE.DEBUG,
                    )

                # TODO: At the moment, this only splits at the compound level (taking the first conformer for resolving),
                if self.config.destination.merge:
                    _write_compounds(self.data.compounds, resource=resource)
                else:
                    for comp in self.data.compounds:
                        _write_compounds([comp], resource)
        elif self.config.compounds.category == _SBE.WRITEOUT_COMP_CATEGORY_ENUMERATIONS:
            if not self.config.destination.format.upper() == _SBE.FORMAT_SDF:
                raise NotImplementedError(
                    "This write-out is not supported for enumerations."
                )
            else:

                def _write_compounds(compounds: List[Compound], resource: str):
                    # TODO: deal with resolving resources differently (also for writing conformers above)
                    resource_resolved = resource
                    for compound in compounds:
                        if len(compound.get_enumerations()) > 0:
                            resource_resolved = resolver.resolve(resource, compounds[0])
                            break

                    self._make_folder(resource_resolved)
                    writer = Chem.SDWriter(resource_resolved)
                    written = 0
                    for comp in compounds:
                        for enum in comp:
                            molecule = enum.get_molecule()
                            if comp.get_name() is not None and comp.get_name() != "":
                                molecule.SetProp(_WE.COMPOUND_NAME, comp.get_name())
                            molecule.SetProp(_WE.RDKIT_NAME, enum.get_index_string())
                            molecule.SetProp(_WE.INDEX_STRING, enum.get_index_string())
                            writer.write(molecule)
                            written += 1
                    writer.close()
                    self._logger.log(
                        f"Wrote {written} enumeration molecules to file {resource_resolved}.",
                        _LE.DEBUG,
                    )

                if self.config.destination.merge:
                    _write_compounds(self.data.compounds, resource=resource)
                else:
                    for comp in self.data.compounds:
                        _write_compounds([comp], resource)
        elif self.config.compounds.category == _SBE.WRITEOUT_COMP_CATEGORY_EXTRADATA:
            if self.config.destination.format.upper() != _SBE.FORMAT_TXT:
                raise ValueError(
                    f"For writing out extra-data (attached to conformers), only TXT is supported as format."
                )
            # TODO: Does merging here makes any sense?
            for comp in self.data.compounds:
                for enum in comp:
                    for conf in enum:
                        resource_resolved = resolver.resolve(resource, conf)
                        self._make_folder(resource_resolved)
                        with open(resource_resolved, "w") as f:
                            content = conf.get_extra_data()[self.config.compounds.key]
                            if isinstance(content, list):
                                for line in content:
                                    f.write(line.rstrip("\n") + "\n")
                            elif isinstance(content, str):
                                f.write(content)
                            else:
                                raise ValueError(
                                    "Extra data must be either a string or a list of strings."
                                )
        else:
            raise ValueError(f"{self.config.compounds.category} not supported.")

    def _write_generic_data(self):
        # type and format do not apply here, simply overwrite defaults
        self.config.destination.type = _SBE.WRITEOUT_DESTINATION_TYPE_FILE
        self.config.destination.format = _SBE.FORMAT_TXT
        resource = self._handle_destination_type()
        self._make_folder(resource)
        if self.config.destination.mode == _SBE.WRITEOUT_DESTINATION_DIR:
            # The output path should be a directory only
            assert not os.path.isfile(resource)
            os.makedirs(resource, exist_ok=True)
        # write out all files from that step with the required extension
        for idx, file in enumerate(
            self.data.generic.get_files_by_extension(self.config.generic.key)
        ):
            if self.config.destination.mode == _SBE.WRITEOUT_DESTINATION_BASE_NAME:
                parts = resource.split(".")
                resource = parts[0] + f"_{idx}." + parts[1]
                file.write(resource, join=False)
            elif self.config.destination.mode == _SBE.WRITEOUT_DESTINATION_AUTOMATIC:
                # take the original file name from the step (these tend not to be very descriptive)
                parts = file.get_file_name().split(".")
                file_name = parts[0] + f"_{idx}." + parts[1]
                resource = os.path.join("/".join(resource.split("/")[:-1]), file_name)
                file.write(resource, join=False)
            elif self.config.destination.mode == _SBE.WRITEOUT_DESTINATION_DIR:
                assert os.path.isdir(resource)
                file.write(resource, join=True, final_writeout=True)

    def _write_gromacs_data(self):
        """
        Handle writeout from gromacs topology state
        """
        self.config.destination.type = _SBE.WRITEOUT_DESTINATION_TYPE_FILE
        self.config.destination.format = _SBE.FORMAT_TXT
        self.config.destination.type = _SBE.WRITEOUT_DESTINATION_DIR
        resource = self._handle_destination_type()
        os.makedirs(resource, exist_ok=True)
        writeout_keys = map(lambda s: s.strip(), self.config.gmx_state.key.split(","))
        for key in writeout_keys:
            if key == _SGE.FIELD_KEY_TOPOL:
                self.data.gmx_state.write_topol(resource)
            elif key == _SGE.FIELD_KEY_NDX:
                self.data.gmx_state.write_ndx(resource)
            elif key == _SGE.PROPS:
                self.data.gmx_state.write_props(resource)
            elif key == _SGE.FIELD_KEY_LOG:
                self.data.gmx_state.write_log(resource)
            elif key == _SGE.FIELD_KEY_TPR:
                self.data.gmx_state.write_tpr(resource)
            elif key == _SGE.FIELD_KEY_CPT:
                self.data.gmx_state.write_cpt(resource)
            elif key == _SGE.FIELD_KEY_EDR:
                self.data.gmx_state.write_edr(resource)
            elif key == _SGE.FIELD_KEY_XTC:
                # if we have multiple trajectories, write them out sequentially, with index attached
                if len(self.data.gmx_state.trajectories.keys()) > 1:
                    for k, v in self.data.gmx_state.trajectories.items():
                        parts = v.get_file_name().split(".")
                        file_name = parts[0] + "_" + str(k) + "." + parts[1]
                        self.data.gmx_state.write_trajectory(
                            resource, file=file_name, index=k
                        )

                else:
                    self.data.gmx_state.write_trajectory(resource)
            elif key == _SGE.FIELD_KEY_STRUCTURE:
                if len(self.data.gmx_state.structures.keys()) > 1:
                    for k, v in self.data.gmx_state.structures.items():
                        parts = v.get_file_name().split(".")
                        file_name = parts[0] + "_" + str(k) + "." + parts[1]
                        self.data.gmx_state.write_structure(
                            resource, file=file_name, index=k
                        )

                else:
                    self.data.gmx_state.write_structure(resource)
            else:
                raise ValueError(
                    f"Gromacs file of type {key} is not supported for writeout"
                )

    def write(self):
        if (
            self.config.compounds is not None
            and self.config.generic is not None
            and self.config.gromacs_state is not None
        ):
            raise ValueError("Only specify one type of writeout per block!")

        if self.config.compounds is not None:
            self._write_compounds()
        elif self.config.generic is not None:
            self._write_generic_data()
        elif self.config.gmx_state is not None:
            self._write_gromacs_data()
        else:
            raise ValueError(
                "Either compounds, generic or gromacs data has to be specified."
            )

    def _writeout_reinvent(self):
        def _get_conf_by_comp_name(confs: List[Conformer], comp_name: str) -> Conformer:
            # assumes there is at most 1 conformer / compound left at this stage, as is required by REINVENT
            for conf in confs:
                if conf.get_compound_name() == comp_name:
                    return conf
            return None

        dict_result = {_WE.JSON_RESULTS: []}
        tags = self._get_selected_tags()

        # add names, including those for which no conformer has been obtained
        dict_result[_WE.JSON_NAMES] = [comp.get_name() for comp in self.data.compounds]

        # do aggregation (might remove conformers)
        confs_unrolled = self._apply_aggregation(self.data.compounds)

        # add values (derived from molecule tags)
        # TODO: if no conformers are left, we need to write out an empty JSON that tells REINVENT that none worked
        for tag in tags:
            values = []
            for comp_name in dict_result[_WE.JSON_NAMES]:
                conf = _get_conf_by_comp_name(confs=confs_unrolled, comp_name=comp_name)
                if conf is not None:
                    try:
                        value = conf.get_molecule().GetProp(tag)
                    except KeyError:
                        value = _WE.JSON_NA
                else:
                    value = _WE.JSON_NA
                values.append(value.strip())
            dict_result[_WE.JSON_RESULTS].append(
                {_WE.JSON_VALUES_KEY: tag, _WE.JSON_VALUES: values}
            )

        # TODO: refactor that part
        resource = self._handle_destination_type()
        if len(confs_unrolled) > 0:
            resolver = RunVariablesResolver()
            resource_resolved = resolver.resolve(resource, confs_unrolled[0])
        else:
            resource_resolved = resource
            self._logger.log(
                f"No conformers obtained, write-out resource resolving disabled.",
                _LE.WARNING,
            )
        self._make_folder(resource_resolved)

        # write-out according to destination type
        # TODO: there seems to be an issue here, when multiple write-out blocks are specified and no conformers are
        #       left: only the first block gets executed and if that's not the REINVENT one, the run will crash
        if self.config.destination.type.lower() in (
            _SBE.WRITEOUT_DESTINATION_TYPE_REINVENT,
            _SBE.WRITEOUT_DESTINATION_TYPE_FILE,
        ):
            with open(resource_resolved, "w") as f:
                json.dump(dict_result, f, indent=4)
        elif self.config.destination.type.lower() in (
            _SBE.WRITEOUT_DESTINATION_TYPE_STDOUT,
            _SBE.WRITEOUT_DESTINATION_TYPE_STDERR,
        ):
            json.dump(dict_result, resource_resolved, indent=4)
        else:
            raise ValueError(
                f"Destination type {self.config.destination.type} not supported for this function."
            )

    def _get_selected_tags(self) -> List[str]:
        # this function returns a list of tags (strings) that are to be considered for e.g. tabular write-out
        # if the respective configuration field is set to "None", use all tags (over all compounds in a batch)
        if self.config.compounds.selected_tags is not None:
            if isinstance(self.config.compounds.selected_tags, list):
                list_tags = self.config.compounds.selected_tags
            elif isinstance(self.config.compounds.selected_tags, str):
                list_tags = [self.config.compounds.selected_tags]
            else:
                raise ValueError(
                    f'Tag selection "{self.config.compounds.selected_tags}" set to illegal value.'
                )
        else:
            # get all tags for all compounds
            list_tags = []
            for comp in self.data.compounds:
                for enum in comp:
                    for conf in enum:
                        list_tags = list_tags + list(conf.get_molecule().GetPropNames())

        list_tags = list(set(list_tags))
        return list_tags

    def _initialize_dict_csv(
        self, keys: List[str], nrow: int, fill_value=np.NaN
    ) -> OrderedDict:
        return_dict = OrderedDict()
        for key in keys:
            return_dict[key] = [fill_value for _ in range(nrow)]
        return return_dict

    def _apply_aggregation(self, compounds: List[Compound]) -> List[Conformer]:
        if (
            self.config.compounds.aggregation.mode
            == _SBE.WRITEOUT_COMP_AGGREGATION_MODE_ALL
        ):
            return self._unroll_conformers(compounds)

        confs_remaining = []
        if (
            self.config.compounds.aggregation.mode
            == _SBE.WRITEOUT_COMP_AGGREGATION_MODE_BESTPERENUMERATION
        ):
            raise NotImplementedError("Best per enumeration is not yet implemented.")
        elif (
            self.config.compounds.aggregation.mode
            == _SBE.WRITEOUT_COMP_AGGREGATION_MODE_BESTPERCOMPOUND
        ):
            for comp in compounds:
                unrolled_conformers = self._unroll_conformers([comp])
                if len(unrolled_conformers) == 0:
                    continue
                values = []
                for conf in unrolled_conformers:
                    try:
                        values.append(
                            float(
                                conf.get_molecule().GetProp(
                                    self.config.compounds.aggregation.key
                                )
                            )
                        )
                    except KeyError as e:
                        self._logger.log(
                            f"Error {e} for conf {conf.get_index_string()}, setting value to 0.00 in writeout!",
                            _LE.WARNING,
                        )
                        values.append(0.00)

                # values = [
                #     float(
                #         conf.get_molecule().GetProp(
                #             self.config.compounds.aggregation.key
                #         )
                #     )
                #     for conf in unrolled_conformers
                # ]
                index_best = (
                    values.index(max(values))
                    if self.config.compounds.aggregation.highest_is_best
                    else values.index(min(values))
                )
                confs_remaining.append(unrolled_conformers[index_best])
        return confs_remaining

    def _unroll_conformers(self, compounds: List[Compound]) -> List[Conformer]:
        result = []
        for comp in compounds:
            for enum in comp:
                for conf in enum:
                    result.append(conf)
        return result

    def _writeout_tabular(self):
        # get all tags of the molecules that are to be considered
        tags = self._get_selected_tags()

        # remove the compound_name and _Name, as they will be specifically added at the beginning
        if _WE.COMPOUND_NAME in tags:
            tags.remove(_WE.COMPOUND_NAME)
        if _WE.RDKIT_NAME in tags:
            tags.remove(_WE.RDKIT_NAME)

        # do aggregation (might remove conformers)
        confs_unrolled = self._apply_aggregation(self.data.compounds)

        # initialize a dictionary with all tags as keys and filled with NA for every position
        dict_result = self._initialize_dict_csv(
            keys=[
                _WE.RDKIT_NAME,
                _WE.COMPOUND_NAME,
                "original_smiles",
                "enumerated_smiles",
            ]
            + tags,
            nrow=len(confs_unrolled),
        )

        # resolve resource
        # TODO: refactor that part
        resource = self._handle_destination_type()
        resolver = RunVariablesResolver()
        if len(confs_unrolled) == 0:
            raise ValueError("No conformers found.")
        resource_resolved = resolver.resolve(resource, confs_unrolled[0])
        self._make_folder(resource_resolved)

        # populate the dictionary with the values (if present)
        for irow in range(len(confs_unrolled)):
            # add the internal Icolos identifier
            conf = confs_unrolled[irow]
            dict_result[_WE.RDKIT_NAME][irow] = conf.get_index_string()

            # add the compound name, if specified
            name = conf.get_compound_name()
            dict_result[_WE.COMPOUND_NAME][irow] = "" if name is None else name
            dict_result["original_smiles"][
                irow
            ] = conf.get_enumeration_object().get_original_smile()

            dict_result["enumerated_smiles"][irow] = Chem.rdmolfiles.MolToSmiles(
                conf.get_molecule()
            )

            for tag in tags:
                try:
                    value = conf.get_molecule().GetProp(tag).strip()
                except KeyError:
                    value = np.nan
                dict_result[tag][irow] = value

        # do the writeout (after sanitation)
        df_result = pd.DataFrame.from_dict(dict_result)
        df_result = self._sanitize_df_columns(df=df_result)
        df_result.to_csv(
            path_or_buf=resource_resolved,
            sep=",",
            na_rep="",
            header=True,
            index=False,
            mode="w",
            quoting=None,
        )
        self._logger.log(
            f"Wrote data frame with {len(confs_unrolled)} rows and {len(tags)} columns to file {resource_resolved}.",
            _LE.DEBUG,
        )

    def _sanitize_df_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        cols_before = df.columns.to_list()
        df.columns = (
            df.columns.str.strip()
            .str.replace(" ", "_")
            .str.replace("(", "")
            .str.replace(")", "")
            .str.replace("/", "_")
            .str.replace("[", "")
            .str.replace("]", "")
        )
        for col_before, col_after in zip(cols_before, df.columns.to_list()):
            if col_before != col_after:
                self._logger.log(
                    f"Sanitized column name {col_before} to {col_after}.", _LE.WARNING
                )
        return df

    def _make_folder(self, path):
        if isinstance(path, str):
            if not os.path.isdir(path):
                path = os.path.dirname(path)
            Path(path).mkdir(parents=True, exist_ok=True)



import abc
import json
import os


class ConfContainer(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, conf):
        # get instance of configuration enum and load configuration
        # parameter "config" can be a string, a path or a dictionary (as long as it holds valid JSON input)
        if isinstance(conf, str):
            if os.path.isfile(conf):
                with open(conf) as file:
                    conf = file.read().replace("\r", "").replace("\n", "")
            conf = json.loads(conf)
        self._conf = conf

    def get_as_dict(self):
        return self._conf

    def get(self, key, default=None):
        return self._conf.get(key, default)

    def __getitem__(self, item):
        return self.get_as_dict()[item]

    def get_as_string(self):
        return json.dumps(self._conf)

    def validate(self):
        raise NotImplementedError(
            "This functions needs to be implemented by child classes."
        )



class TurbomoleEnum:

    # general
    # ---------
    COORD = "coord"  # hard-coded file name of input coordinates
    CONTROL = "control"  # hard-coded file name for control script
    TM_CONFIG_DIR = "tm_config_dir"  # directory path where the *.tm configurations lie
    # the basename of the parameter set chosen, e.g. "b97-3c-ri-d3-def2-mtzvp-int-nosym-charge"
    TM_CONFIG_BASENAME = "tm_config_basename"
    # this does not contain the charge or the ending
    # full name e.g. "b97-3c-ri-d3-def2-mtzvp-int-nosym-charge-1.tm"
    TM_CONFIG_ENDING = ".tm"  # ending of turbomole configuration files
    TM_CONFIG_COSMO = "tm_config_cosmo"  # path to the COSMO configuration file
    # this needs to be set for each turbomole calculation (current folder)
    TM_TURBOTMPDIR = "TURBOTMPDIR"
    TM_OUTPUT_COSMOFILE = "mol.cosmo"  # hard-coded file name of the turbomole output
    TM_OUTPUT_COORDFILE = "coord"
    TM_OUTPUT_FINAL_XYZ = "final.xyz"  # hard-coded file name for turbomole output
    TM_OUTPUT_FINAL_SDF = "final.sdf"

    # the "ridft" binary (+ configuration elements)
    # ---------
    TM_RIDFT = "ridft"  # do DFT calculation with RI-J approximation for
    # inter-electronic Coulomb term
    TM_RIDFT_FAIL_IDENTIFICATION_STRING = "ridft ended abnormally"
    TM_RIDFT_SUCCESS_STRING = (
        "ridft ended normally"  # if this string is in stderr, execution was successful
    )

    # the "jobex" binary (+ configuration elements)
    # ---------
    TM_JOBEX = "jobex"  # used for DFT optimization
    TM_JOBEX_C = "-c"
    TM_JOBEX_GCART = "-gcart"
    TM_JOBEX_FAIL_IDENTIFICATION_STRING = "jobex ended abnormally"
    TM_JOBEX_SUCCESS_STRING = "jobex ended normally"

    # the "cosmoprep" binary (+ configuration elements)
    # ---------
    TM_COSMOPREP = "cosmoprep"
    TM_COSMOPREP_SUCCESS_STRING = "cosmoprep ended normally"

    # the "define" binary (+ configuration elements)
    # ---------
    TM_DEFINE = "define"
    TM_DEFINE_SUCCESS_STRING = (
        "define ended normally"  # if this string is in stderr, execution was successful
    )

    # the "x2t" binary (+ configuration elements)
    # ---------
    TM_X2T = (
        "x2t"  # program to translate an XYZ file to TM input: x2t input.xyz > coord
    )
    TM_X2T_SUCCESS_STRING = (
        "$coord"  # if this string is in stdout, execution was successful
    )

    # the "t2x" binary (+ configuration elements)
    # ---------
    # program to translate an TM input to an XYZ file (last snapshot): t2x -c > final.xyz
    TM_T2X = "t2x"
    TM_T2X_C = "-c"

    # the "cosmotherm" binary (+ configuration elements)
    # ---------
    CT_COSMOTHERM = "cosmotherm"  # the cosmotherm binary
    CT_COSMOTHERM_FAIL_STRING = (
        "COSMOtherm ERROR Termination"  # if this string is in stderr, the job failed
    )
    CT_COSMOTHERM_CONFIG_FILE = (
        "cosmotherm.inp"  # hard-coded name of the input file generated before execution
    )
    CT_COSMOTHERM_OUTPUT_FILE = (
        "cosmotherm.out"  # hard-coded name of the output file generated by cosmotherm
    )
    CT_COSMOTHERM_TAB_ENDING = "cosmother.tab"

    CT_CONFIG = "cosmotherm_config"
    CT_CONFIG_DEFAULTPATH = "src/icolos/config/cosmo/default_cosmo.config"

    # control script fields
    # ---------
    CONTROL_COSMO_OUT = (
        # line, after which insertion is to be put (separate line)
        "$cosmo_out file=n"
    )
    CONTROL_COSMO_INSERTION = "$cosmo_isorad"
    CONTROL_COSMO_REPLACE = (
        "$cosmo_out file=mol.cosmo"  # after insertion, replace "$cosmo_out" with this
    )

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class CosmoOutputEnum:

    PATTERN = "pattern"
    ELEMENT = "element"

    # general block
    # ---------
    # this works, because we always set the name of the compound to "mol"
    GENERAL_BLOCK_PATTERN_STRING = "--- Compound 1 (mol) ---"
    # alternatively, one could search for "Compound 1 ", as we also feed it in in order

    # the key (e.g. "E_cosmo") is the tag name for the SDF write-out, "pattern" identifies the line and
    # "element" is the number (on the right side, after the ':') of the element (starting with 0) from a
    # split using ' ', respectively
    GENERAL_BLOCK_ANNOTATIONS = {
        "E_cosmo": {"pattern": "E_COSMO+dE", "element": 0},
        "volume": {"pattern": "Volume", "element": 0},
        "area": {"pattern": "Area", "element": 0},
        "dipole": {"pattern": "Dipole moment", "element": 0},
        "HB_acc": {"pattern": "H-bond moment (accept)", "element": 0},
        "HB_don": {"pattern": "H-bond moment (donor)", "element": 0},
        "sigma1": {"pattern": "Sigma moments", "element": 0},
        "sigma2": {"pattern": "Sigma moments", "element": 1},
        "sigma3": {"pattern": "Sigma moments", "element": 2},
        "sigma4": {"pattern": "Sigma moments", "element": 3},
        "sigma5": {"pattern": "Sigma moments", "element": 4},
        "sigma6": {"pattern": "Sigma moments", "element": 5},
    }

    # solvent blocks
    # ---------
    SOLVENT_BLOCK_PATTERN_STRING = "Gibbs Free Energy of Solvation"
    SOLVENT_BLOCK_START_PATTERN = "----------------------"
    SOLVENT_BLOCK_BODY_START_PATTERN = "Compound: 1  (mol)"
    SOLVENT_TRANSLATE_SOLVENT = {
        "h2o": "h2o",
        "methanol": "meoh",
        "1-octanol": "octanol",
        "dimethyls": "dmso",
        "cyclohexa": "cychex",
        "chcl3": "chcl3",
        "acetonitr": "acn",
        "thf": "thf",
    }
    SOLVENT_REPLACEHOLDER = "{solvent}"
    SOLVENT_BLOCK_BODY_ANNOTATIONS = {
        "Gsolv_{solvent}": {"pattern": "Gibbs Free Energy of Solvation", "element": 0},
        "G_{solvent}": {"pattern": "Free energy of molecule in mix", "element": 0},
    }
    SOLVENT_BLOCK_HEADER_COMPOUNDS_PATTERN = "Compound   "
    SOLVENT_BLOCK_HEADER_MOLFRACTION_PATTERN = "Mole Fraction"
    SOLVENT_BLOCK_CURRENT_FRACTION_VALUE = "1.0000"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class FeatureCounterEnum:

    PROPERTY_NUM_RINGS = "num_rings"
    PROPERTY_NUM_AROMATIC_RINGS = "num_aromatic_rings"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class CrestEnum:

    # Note: The first argument is usually a coordinate file in TM (coord, Bohr),
    #       Xmol (*.xyz, Ang.) or SDF format.
    # Call: "crest <input_file> <parameters>"
    # This collection is based on Version 2.10.2, compatible with XTB version 6.1 and later

    # General options (all)
    # ---------
    CREST = "crest"  # binary name
    CREST_H = "-h"  # print the help message
    CREST_HELP_IDENTIFICATION_STRING = "Conformer-Rotamer Ensemble Sampling Tool"
    CREST_V3 = "-v3"  # version 3 (the default iMTD-GC workflows)
    CREST_G = "-g"  # 1 string parameter; use GBSA implicit solvent for solvent <string>
    CREST_CHRG = "-chrg"  # 1 int parameter; the molecule's charge
    CREST_UHF = "-uhf"  # 1 int parameter; set <int>=Nα-Nβ electrons
    CREST_NOZS = (
        "-nozs"  # do not perform z-mat sorting, default: z-matrix will be sorted
    )
    CREST_ZS = "-zs"  # perform z-matrix sorting [default]
    # 1 level parameter (vloose, loose, normal, tight, vtight); default: vtight
    CREST_OPT = "-opt"
    CREST_GFN1 = "-gfn1"  # use GFN1-xTB
    CREST_GFN2 = "-gfn2"  # use GFN2-xTB [default]
    CREST_GFF = "-gff"  # use GFN-FF (requires xtb 6.3 or newer)
    # 1 string parameter; specify name of the xtb binary that should be used
    CREST_XNAM = "-xnam"
    # 1 float parameter; set energy window in kcl/mol, default: 6.0 kcal/mol
    CREST_EWIN = "-ewin"
    CREST_RTHR = (
        "-rthr"  # 1 float parameter; set RMSD threshold in Ang, default: 0.125 Ang
    )
    CREST_ETHR = (
        "-ethr"  # 1 float parameter; set E threshold in kcal/mol, default: 0.1 kcal/mol
    )
    CREST_BTHR = (
        "-bthr"  # 1 float parameter; set Rot. const. threshold, default: 15.0 MHz
    )
    # 1 float parameter; Boltzmann population threshold, default: 0.05 (= 5%)
    CREST_PTHR = "-pthr"
    CREST_EQV = "-eqv"  # activate NMR-equivalence printout
    CERST_NMR = "-nmr"  # activate NMR-mode (= [-eqv] + opt. level: vtight)
    CREST_PRSC = "-prsc"  # create a scoord.* file for each conformer
    CREST_NICEPRINT = "-niceprint"  # progress bar printout for optimizations
    CREST_DRY = "-dry"  # performs a "dry run"; only prints the settings

    # iMTD-GC workflows (selected)
    # ---------
    CREST_CROSS = "-cross"  # do the GC part [default]
    CREST_NOCROSS = "-nocross"  # don't do the GC part
    # 1 int parameter; set SHAKE mode for MD (0=off, 1=H-only, 2=all bonds), default: 2
    CREST_SHAKE = "-shake"
    CREST_TSTEP = "-tstep"  # 1 int parameter; set MD time step in fs, default: 5

    # other (selected)
    # ---------
    CREST_T = (
        # 1 int parameter; set total compound_number of CPUs (threads) to be used
        "-T"
    )

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class AutoDockVinaEnum:
    # executable "vina" + parameters
    # ---------
    VINA = "vina"
    VINA_CALL = "vina"  # the binary call
    VINA_HELP = "--help"  # display usage summary
    VINA_HELP_ADVANCED = "--help_advanced"  # display usage summary (with all options)
    VINA_VERSION = "--version"  # diplay program version
    VINA_VERSION_IDENTIFICATION_STRING = (
        "AutoDock Vina 1.1.2"  # string, which needs to be present in help output in
    )
    # order to assume "AutoDock Vina" can be properly used
    VINA_CONFIGURATION = (
        "--config"  # path to configuration file, where options below can be put
    )

    # input
    VINA_RECEPTOR = "--receptor"  # rigid part of the receptor (PDBQT)
    VINA_LIGAND = "--ligand"  # ligand (PDBQT); only one at a time
    VINA_FLEX = "--flex"  # flexible side chains, if any (PDBQT)

    # search space
    VINA_CENTER_X = "--center_x"  # X coordinate of the center
    VINA_CENTER_Y = "--center_y"  # Y coordinate of the center
    VINA_CENTER_Z = "--center_z"  # Z coordinate of the center
    VINA_SIZE_X = "--size_x"  # size in the X dimension (Angstroms)
    VINA_SIZE_Y = "--size_y"  # size in the X dimension (Angstroms)
    VINA_SIZE_Z = "--size_z"  # size in the X dimension (Angstroms)

    # output
    VINA_OUT = "--out"  # output models (PDBQT), the default is chosen based on the
    # ligand file name

    # advanced options
    VINA_SCORE_ONLY = "--score_only"  # score only - search space can be omitted
    VINA_LOCAL_ONLY = "--local_only"  # do local search only
    VINA_RANDOMIZE_ONLY = (
        "--randomize_only"  # randomize input, attempting to avoid clashes
    )
    VINA_WEIGHT_GAUSS1 = "--weight_gauss1"  # gauss_1 weight (default: -0.035579)
    VINA_WEIGHT_GAUSS2 = "--weight_gauss2"  # gauss_2 weight (default: -0.005156)
    VINA_WEIGHT_REPULSION = (
        "--weight_repulsion"  # repulsion weight (default: 0.84024500000000002)
    )
    VINA_WEIGHT_HYDROPHOBIC = (
        "--weight_hydrophobic"  # hydrophobic weight (-0.035069000000000003)
    )
    VINA_WEIGHT_HYDROGEN = (
        "--weight_hydrogen"  # hydrogen bond weight (-0.58743900000000004)
    )
    VINA_WEIGHT_ROT = "--weight_rot"  # N_rot weight (default: 0.058459999999999998)

    # miscellaneous (optional)
    VINA_CPU = "--cpu"  # the number of CPUs to use (the default is to try to detect
    # the number of CPUs or, failing that, use 1)
    VINA_SEED = "--seed"  # explicit random seed
    VINA_EXHAUSTIVENESS = (
        "--exhaustiveness"  # exhaustiveness of the global search (roughly proportional
    )
    # to time): 1+ (default: 8)
    VINA_NUM_MODES = (
        "--num_modes"  # maximum number of binding modes to generate (default: 9)
    )
    VINA_ENERGY_RANGE = "--energy_range"  # maximum energy difference between the best binding mode and the
    # worst one displayed [kcal/mol] (default: 3)

    # ---------
    # Vina output specifications
    # ---------
    ADV_PDBQT = ".pdbqt"

    # the score is part of a tag in the PDBQT -> SDF translated output (tag "REMARK"), which looks like that:
    # < REMARK >
    # VINA RESULT: -9.1 0.000 0.000
    # Name = /tmp/tmpjssiy8z4.pdb
    # ...

    # Note, that the three values are: affinity [kcal/mol] | dist from best mode (rmsd l.b.) | rmsd (u. b.)
    REMARK_TAG = "REMARK"
    RESULT_LINE_IDENTIFIER = "VINA RESULT"
    RESULT_LINE_POS_SCORE = 2
    RESULT_LINE_POS_RMSDTOBEST_LB = 3
    RESULT_LINE_POS_RMSDTOBEST_UB = 4

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class GoldEnum:

    GOLD_CALL = "gold_auto"
    GOLD_HELP = "-h"
    GOLD_HELP_IDENTIFICATION_STRING = "Usage: gold_auto"
    GOLD_QUIET = "-q"

    REMARK_TAG = "REMARK"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class GoldOutputEnum:

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class CrestOutputEnum:

    COORD = "coord"
    COORD_ORIGINAL = "coord.original"
    CRE_MEMBERS = "cre_members"
    CREST_ENERGIES = "crest.energies"
    CREST_BEST_XYZ = "crest_best.xyz"
    CREST_CONFORMERS_SDF = "crest_conformers.sdf"
    CREST_CONFORMERS_XYZ = "crest_conformers.xyz"
    CREST_ROTAMERS_XYZ = "crest_rotamers.xyz"
    XTBTOPO_MOL2 = "xtbtopo.mol"

    # format properties
    PREFIX_ENERGIES_XYZ = "        "

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class OpenBabelEnum:

    # executable "obabel" + parameters
    # ---------
    OBABEL = "obabel"
    OBABEL_IDENTIFICATION_STRING = "-O<outfilename>"
    OBABEL_INPUTFORMAT_PDBQT = (
        # sets the input format to "PDBQT" (output of "AutoDock Vina")
        "-ipdbqt"
    )
    OBABEL_INPUTFORMAT_XYZ = (
        "-ixyz"  # sets the input format to "XYZ" (format in XTB/TM)
    )
    OBABEL_INPUTFORMAT_PDB = "-ipdb"
    OBABEL_INPUTFORMAT_SDF = "-isdf"  # sets the input format to "SDF"
    OBABEL_P = "-p"  # sets the <pH> value (e.g. "-p 7.4") for protonation
    # note, that this overwrites "--addpolarh", which is thus not used
    # specifies the output path (directly pasted afterwards, e.g. "-Omypath.pdb")
    OBABEL_O = "-O"
    # sets the output format to "PDBQT" (input for "AutoDock Vina")
    OBABEL_OUTPUT_FORMAT_PDBQT = "-opdbqt"
    OBABEL_OUTPUT_FORMAT_SDF = "-osdf"  # sets the output format to "SDF"
    # sets the output format to "XYZ" (format in XTB/TM)
    OBABEL_OUTPUT_FORMAT_XYZ = "-oxyz"
    OBABEL_OUTPUT_FORMAT_PDB = "-opdb"
    OBABEL_X = "-x"  # specifies generation options
    OBABEL_M = "-m"  # produce multiple output files
    # one of the 'X' options ("-x"), which disables the tree construction of the receptor
    # (makes it static), directly pasted together: e.g. "-xr"
    OBABEL_X_R = "r"
    # sets the partial charge generation method (execute "obabel -L charges" to see list of available methods)
    OBABEL_PARTIALCHARGE = "--partialcharge"
    # one method to compute the partial charges, used as: "--partialcharge gasteiger"
    OBABEL_PARTIALCHARGE_GASTEIGER = "gasteiger"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class OpenBabelOutputEnum:

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class OMEGAEnum:

    # executable "oeomega" + parameters; the first parameter is a string indicating the mode
    # ---------
    OMEGA = "oeomega"
    OMEGA_HELP = "--help"  # print the help message
    OMEGA_HELP_IDENTIFICATION_STRING = "To cite OMEGA please"
    OMEGA_MODE_CLASSIC = "classic"  # The original customizable omega2 interface
    OMEGA_MODE_MACROCYCLE = "macrocycle"  # Conformer generation for macrocycles
    OMEGA_MODE_ROCS = "rocs"  # Optimal conformer generation for ROCS
    OMEGA_MODE_POSE = "pose"  # Optimal conformer generation for molecular
    # alignment and pose prediction by docking
    OMEGA_MODE_DENSE = "dense"  # Optimal conformer generation for FREEDOM

    # mode "classic" parameters
    # ---------
    CLASSIC_INPUT = "-in"  # Input filename (required, if "-param" not set)
    CLASSIC_OUTPUT = "-out"  # Output filename (required, if "-param" not set)
    CLASSIC_PARAM = "-param"  # A parameter file
    CLASSIC_PREFIX = "-prefix"  # Prefix to use to name output files
    CLASSIC_PROGRESS = "-progress"  # Method of showing job progress. Either "none",
    # "dots", "log" or "percent".
    CLASSIC_SDENERGY = "-sdEnergy"  # Writes conformer energies to the SD tag field
    CLASSIC_VERBOSE = "-verbose"  # Triggers copious logging output
    # Generate structures from connection-table only.
    CLASSIC_FROMCT = "-fromCT"
    CLASSIC_EWINDOW = "-ewindow"  # Energy window used for conformer selection.
    CLASSIC_MAXCONFS = (
        "-maxconfs"  # Maximum compound_number of conformations to be saved
    )
    CLASSIC_RMS = "-rms"  # RMS threshold used to determine duplicate
    # conformations
    # if set to false ("-canonOrder false"), OMEGA will not update the atom orders
    CLASSIC_CANON_ORDER = "-canonOrder"
    CLASSIC_STRICTSTEREO = (
        "-strictstereo"  # Requires that all chiral atoms and bonds have
    )
    # specified stereo
    CLASSIC_STRICT = "-strict"  # A convenience flag to set "-strictstereo",
    # "-strictatomtyping" and "-strictfrags" to true
    # or false and override [sic] previous settings.

    # mode "rocs" parameters
    # ---------
    ROCS_INPUT = "-in"  # Input filename (required, if "-param" not set)
    ROCS_OUTPUT = "-out"  # Output filename (required, if "-param" not set)
    ROCS_PARAM = "-param"  # A parameter file
    ROCS_PREFIX = "-prefix"  # Prefix to use to name output files
    ROCS_PROGRESS = "-progress"  # Method of showing job progress. Either "none",
    # "dots", "log" or "percent".
    ROCS_VERBOSE = "-verbose"  # Triggers copious logging output

    # mode "dense" parameters
    # ---------
    DENSE_INPUT = "-in"  # Input filename (required, if "-param" not set)
    DENSE_OUTPUT = "-out"  # Output filename (required, if "-param" not set)
    DENSE_PARAM = "-param"  # A parameter file
    DENSE_PREFIX = "-prefix"  # Prefix to use to name output files
    DENSE_PROGRESS = "-progress"  # Method of showing job progress. Either "none",
    # "dots", "log" or "percent".
    DENSE_VERBOSE = "-verbose"  # Triggers copious logging output

    # mode "macrocycle" parameters
    # ---------
    MACROCYCLE_INPUT = "-in"  # Input filename (required, if "-param" not set)
    # Output filename (required, if "-param" not set)
    MACROCYCLE_OUTPUT = "-out"
    MACROCYCLE_PARAM = "-param"  # A parameter file
    MACROCYCLE_PREFIX = "-prefix"  # Prefix to use to name output files
    MACROCYCLE_EWINDOW = "-ewindow"  # Energy window for the output conformers
    MACROCYCLE_ITERATION_CYCLE_SIZE = (
        "-iteration_cycle_size"  # Number of iterations to run before checking if a
    )
    # new minimum was found (run will finish if no new
    # minimum is found).
    MACROCYCLE_MAXCONFS = (
        "-maxconfs"  # Maximum compound_number of conformations to be saved
    )
    MACROCYCLE_MAX_ITERATIONS = (
        # Maximum compound_number of iterations (calculation may
        "-max_iterations"
    )
    # converge before reaching this compound_number).
    MACROCYCLE_REF_TOLERANCE = (
        "-ref_tolerance"  # RMS gradient tolerance for force field refinement
    )
    MACROCYCLE_RMS = "-rms"  # RMS clustering threshold (if 0.0 clustering is
    # skipped).
    MACROCYCLE_RMSD_DEDUPLICATE = (
        "-rmsd_deduplicate"  # Deduplicate using a RMSD calculation (slow)
    )
    # rather than energy and torsion comparison

    # mode "pose" parameters
    # ---------
    POSE_INPUT = "-in"  # Input filename (required, if "-param" not set)
    POSE_OUTPUT = "-out"  # Output filename (required, if "-param" not set)
    POSE_PREFIX = "-prefix"  # Prefix to use to name output files
    POSE_PROGRESS = "-progress"  # Method of showing job progress. Either "none",
    # "dots", "log" or "percent".
    POSE_VERBOSE = "-verbose"  # Triggers copious logging output

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class OMEGAOutputEnum:

    # tags
    CLASSIC_ENERGY_OUTPUT_TAG = "mmff94smod_NoEstat"

    # other
    # This hard-coded output name will be parsed.
    OUTPUT_SDF_NAME = "omega_out.sdf"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class XTBEnum:

    # Usage: xtb[options] <geometry> [options]
    # < geometry > may be provided as valid TM coordinate file(*coord in Bohr) or in xmol format(*xyz in Ångström).
    # Output Conventions: total energies are given in atomic units (Eh), gaps/HL energies are given in eV
    XTB = "xtb"
    XTB_HELP = "--help"
    XTB_HELP_IDENTIFICATION_STRING = "normal termination of xtb"  # written to stderr
    XTB_CHRG = "--chrg"  # 1 int parameter; specify molecular charge
    XTB_UHF = "--uhf"  # 1 int parameter; specify Nalph-Nbeta
    # 1 float parameter; accuracy for SCC calculation, lower is better (default = 1.0)
    XTB_ACC = "--acc"
    # 1 int parameter; compound_number of iterations in SCC (default = 250)
    XTB_ITERATION = "--iteration"
    # 1 level parameter; compound_number of cycles in ANCopt (default = automatic)
    XTB_CYCLES = "--cycles"
    XTB_GFN = (
        # 1 int parameter; specify parametrisation of GFN-xTB (default = 2)
        "--gfn"
    )
    XTB_QMDFF = "--qmdff"  # use QMDFF for single point (needs solvent-file)
    XTB_TM = "--tm"  # use TURBOMOLE for single point (needs control-file)
    XTB_ORCA = "--orca"  # use ORCA for single point (writes ORCA input)
    XTB_MOPAC = "--mopac"  # use MOPAC for single point (writes MOPAC input)
    # uses periodic boundary conditions (in developement)
    XTB_PERIODIC = "--periodic"
    # 1 float parameter; electronic temperature (default = 300K)
    XTB_ETEMP = "--etemp"
    # 1 level parameter; generalized born (GB) model with solvent accessable surface area (SASA) model
    XTB_GBSA = "--gbsa"
    XTB_OPT = "--opt"  # 1 level parameter; either "crude", "sloppy",
    # "loose", "normal" (default), "tight", "verytight"
    XTB_P = "-P"  # 1 int parameter; compound_number of cores

    # --vparam FILE  Parameter file for vTB calculation
    # --xparam FILE  Parameter file for xTB calculation (not used)
    # --pop          requests printout of Mulliken population analysis
    # --molden       requests printout of molden file
    # --dipole      requests dipole printout
    # --wbo          requests Wiberg bond order printout
    # --lmo          requests localization of orbitals
    # --fod          requests FOD calculation, adjusts electronic temperature to 12500 K if possible
    # --scc, --sp    performs a single point calculation
    # --vip          performs calculation of ionisation potential
    # --vea          performs calculation of electron affinity
    # --vipea        performs calculation of IP and EA
    # --vomega       performs calculation of electrophilicity index
    # --vfukui       calculate Fukui indicies using GFN-xTB
    # --esp          calculate electrostatic potential on VdW-grid
    # --stm          calculate STM image
    # --grad         performs a gradient calculation
    # --optts [LEVEL] [ROOT] call ancopt(3) to perform a transition state optimization, may
    # need to perform a hessian calculation first
    # --hess         perform a numerical hessian calculation on input geometry
    # --ohess [LEVEL] perform a numerical hessian calculation on an ancopt(3) optimized geometry
    # --md           molecular dynamics simulation on start geometry
    # --omd          molecular dynamics simulation on ancopt(3) optimized geometry, a loose
    # optimization level will be chosen.
    # --metadyn [INT] meta dynamics simulation on start geometry saving INT snapshots to bias the simulation
    # --siman        conformational search by simulated annealing based on molecular dynamics.
    # Conformers are optimized with ancopt.
    # --modef INT    modefollowing algorithm.  INT specifies the mode that should be used for the modefollowing.
    # -I,--input FILE   use FILE as input source for xcontrol(7) instructions
    # --namespace STRING give this xtb(1) run a namespace. All files, even temporary ones, will
    # be named accordingly (might not work everywhere).
    # --[no]copy     copies the xcontrol file at startup (default = true)
    # --[no]restart  restarts calculation from xtbrestart (default = true)
    # -P,--parallel INT compound_number of parallel processes
    # --define       performs automatic check of input and terminate
    # --version      print version and terminate
    # --citation     print citation and terminate
    # --license      print license and terminate
    # -v,--verbose      be more verbose (not supported in every unit)

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class XTBOutputEnum:

    XTBOPT_SDF = "xtbopt.sdf"
    XTBTOPO_SDF = "xtbtopo.sdf"
    XTBOPT_LOG = "xtbopt.log"
    XTBRESTART = "xtbrestart"
    WBO = "wbo"
    CHARGES = "charges"
    SUCCESS = "success"
    FAILURE = "failure"

    # tags
    TOTAL_ENERGY_TAG = "total energy / Eh"
    GRADIENT_TAG = "gradient norm / Eh/a0"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class MacromodelEnum:

    MACROMODEL = "macromodel"
    MACROMODEL_HELP = "-h"
    MACROMODEL_HELP_IDENTIFICATION_STRING = "MacroModel Startup Script"
    MACROMODEL_NJOBS = "-NJOBS"
    MACROMODEL_WAIT = "-WAIT"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class ModelBuilderEnum:

    # OPTBUILD parameters
    OPTBUILD_ENTRY_POINT = "optbuild.py"
    CONFIG = "--config"
    BEST_BUILDCONFIG_OUTPATH = (
        "--best-buildconfig-outpath"  # path to the output JSON for the best trial
    )
    BEST_MODEL_OUTPATH = (
        # path to the output model (PKL) for the best trial
        "--best-model-outpath"
    )
    MERGED_MODEL_OUTPATH = (
        "--merged-model-outpath"  # path to the production output model (PKL)
    )
    PERSISTENCE_MODE = "--model-persistence-mode"
    PERSISTENCE_MODE_PLAINSKLEARN = "plain_sklearn"
    PERSISTENCE_MODE_SKLEARNWITHOPTUNAAZ = "sklearn_with_optunaz"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class LigprepEnum:

    LIGPREP = "ligprep"
    LIGPREP_HELP = "-h"
    LIGPREP_HELP_IDENTIFICATION_STRING = "usage: ligprep [options]"
    # SMI input followed by <path> (alternatives: "-icsv", "-imae" and "-isd")
    LIGPREP_INPUT_ISMI = "-ismi"
    LIGPREP_OUTPUT_OSD = (
        "-osd"  # SD(F) output followed by <path> (alternative: "-omae")
    )
    # not used in AZdock, but would be an option to feed parameters from configuration file
    LIGPREP_INP_CONFIG = "-inp"
    LIGPREP_EPIK = (
        "-epik"  # Use "Epik" for ionization and tautomerization (Recommended)
    )
    LIGPREP_PH = (
        # Effective / target pH; followed by <number> (use 7.0 as default)
        "-ph"
    )
    # pH tolerance for generated structures; followed by <number> (use 2.0 as default)
    LIGPREP_PHT = "-pht"
    LIGPREP_AC = (
        "-ac"  # Do not respect existing chirality properties and do not respect
    )
    # chiralities from the input geometry. Generate stereoisomers for all chiral centers up to
    # the number permitted (specified using the -s option). This is equivalent to "Generate
    # all combinations" in the Ligand Preparation user interface. Default
    # behavior is to respect only explicitly indicated chiralities.
    # Filter structures via LigFilter using specifications from the file provided. Default: do not filter.
    LIGPREP_F = "-f"
    LIGPREP_G = (
        # Respect chiralities from input geometry when generating stereoisomers.
        "-g"
    )
    # Generate up to this <number> stereoisomers per input structure. (Default: 32).
    LIGPREP_S = "-s"
    # Force-field to be used for the final geometry optimization (either 14 or 16, which refers to OPLS_2005 and
    LIGPREP_BFF = "-bff"
    # OPLS3e respectively. Default: 14
    LIGPREP_FF_OPLS_2005 = "14"  # Default force-field
    LIGPREP_FF_OPLS3e = "16"  # Alternative force-field
    LIGPREP_NJOBS = (
        # Divide the overall job into NJOBS subjobs. Set to 1 by default.
        "-NJOBS"
    )
    # Divide the overall job into subjobs with no more than NSTRUCTS structures. Set to 1 by default.
    LIGPREP_NSTRUCTS = "-NSTRUCTS"
    LIGPREP_HOST = (
        # Run the job on <hostname> remotely on the indicated host entry.
        "-HOST"
    )
    LIGPREP_HOST_LOCALHOST = "localhost"  # Default value for the run.
    LIGPREP_WAIT = "-WAIT"  # Do not return a prompt until the job completes.

    LIGPREP_LOG_ENDING = ".log"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class GlideEnum:
    # executable "glide" + parameters
    # note, that you can get the full list of parameters with "$SCHRODINGER/glide -k"
    # ---------
    GLIDE = "glide"
    GLIDE_CALL = "$SCHRODINGER/glide"
    GLIDE_HELP = "-h"
    GLIDE_HELP_IDENTIFICATION_STRING = "positional arguments:"
    GLIDE_WAIT = "-WAIT"
    GLIDE_OVERWRITE = "-OVERWRITE"  # Remove previous job files before running.
    GLIDE_NJOBS = "-NJOBS"  # Divide the overall job into NJOBS subjobs.
    GLIDE_HOST = "-HOST"  # Run job remotely on the indicated host entry.
    # WARNING: does not seem to be supported (any longer?) - probably "-NOLOCAL" now?
    GLIDE_TMPLAUNCHDIR = "-TMPLAUNCHDIR"
    # WARNING: does not seem to be supported (any longer?)
    GLIDE_ATTACHED = "-ATTACHED"
    # amide bond rotation behavior: "fixed", "free", "penal", "trans", "gen[eralized]"
    GLIDE_AMIDE_MODE = "AMIDE_MODE"
    # bypass elimination of poses in rough scoring stage (useful for fragment docking)
    GLIDE_EXPANDED_SAMPLING = "EXPANDED_SAMPLING"
    GLIDE_GRIDFILE = "GRIDFILE"  # path to grid (.grd or .zip) file
    GLIDE_LIGANDFILE = "LIGANDFILE"  # Glide docking ligands file name
    # expand size of the Glide funnel by N times to process poses from N confgen runs with minor
    # perturbations to the input ligand coordinates
    GLIDE_NENHANCED_SAMPLING = "NENHANCED_SAMPLING"
    # format for file containing docked poses: "poseviewer" for _pv.mae output; "ligandlib" for
    # _lib.mae; similarly "poseviewer_sd" and "ligandlib_sd" for sdf output; "phase_subset" for bypassing
    # _lib or _pv in favor of a Phase subset file.
    GLIDE_POSE_OUTTYPE = "POSE_OUTTYPE"
    GLIDE_POSE_OUTTYPE_LIGANDLIB = (
        "ligandlib_sd"  # sets the write-out to SDF (easily parsed)
    )
    # uses the poseviewer (MAE format) write-out; contains the receptor
    GLIDE_POSE_OUTTYPE_POSEVIEWER = "poseviewer"
    GLIDE_POSES_PER_LIG = (
        "POSES_PER_LIG"  # maximum number of poses to report per each input ligand
    )
    # maximum number of best-by-Emodel poses to submit to post-docking minimization
    GLIDE_POSTDOCK_NPOSE = "POSTDOCK_NPOSE"
    GLIDE_POSTDOCKSTRAIN = (
        "POSTDOCKSTRAIN"  # include strain correction in post-docking score
    )
    # glide docking precision ("SP", "XP" or "HTVS")
    GLIDE_PRECISION = "PRECISION"
    # reward formation of intramolecular hydrogen bonds in the ligand
    GLIDE_REWARD_INTRA_HBONDS = "REWARD_INTRA_HBONDS"
    GLIDE_USE_CONS = "USE_CONS"
    GLIDE_NREQUIRED_CONS = "NREQUIRED_CONS"
    # if any of these string is present in the logfile associated with a subjob, all went well
    GLIDE_LOG_SUCCESS_STRING = "glide_sort command succeeded"
    GLIDE_LOG_FINISHED_STRINGS = {"Exiting Glide"}
    GLIDE_LOG_FAIL_STRINGS = {
        "*** Error in",
        # if any of these strings is present in the logfile associated with a subjob, there was an
        # issue resulting in the complete failure of the execution
        "Glide cannot recover from this signal and will now abort.",
        "======= Backtrace: =========",
    }

    # "Glide: FATAL mmlewis error"}

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class SchrodingerExecutablesEnum:

    # executable "licadmin" + parameters
    # ---------
    LICADMIN = "licadmin"
    LICADMIN_STAT = "STAT"  # returns the list of tokens used / available

    # executable "sdconvert" + parameters
    # ---------
    SCHRODINGER_MODULE = "module load schrodinger/2021-2-js-aws"
    SDCONVERT = "sdconvert"
    SDCONVERT_CALL = "$SCHRODINGER/utilities/sdconvert"
    SDCONVERT_HELP = ""
    SDCONVERT_HELP_IDENTIFICATION_STRING = "mae : Maestro format"
    SDCONVERT_A = "-a"  # append structures to the output file
    # input; note that the format is directly appended (e.g. "-isd")
    SDCONVERT_I = "-i"
    SDCONVERT_O = (
        # output; note that the format is directly appended (e.g. "-omae")
        "-o"
    )
    SDCONVERT_FORMAT_SD = "sd"  # MDL SDfile format
    SDCONVERT_FORMAT_PDB = "pdb"  # PDB file format
    SDCONVERT_FORMAT_MM = "mm"  # MacroModel (.dat) format
    SDCONVERT_FORMAT_MAE = "mae"  # Maestro format
    SDCONVERT_TITLE = (
        "-title"  # define SD property <prop> as the source of the Maestro title
    )
    SDCONVERT_NOSTEREO = (
        "-nostereo"  # do not record the atom parity info from the input file
    )
    # do not convert aromatic type 4 bonds to single and double bonds (which is the Maestro convention)
    SDCONVERT_NOAROM = "-noarom"

    # executable "structcat" + parameters
    STRUCTCAT = "structcat"
    STRUCT_SPLIT_CALL = "$SCHRODINGER/run split_structure.py"
    STRUCT_SPLIT = "structsplit"
    STRUCTCONVERT = "structconvert"
    STRUCTCAT_CALL = "$SCHRODINGER/utilities/structcat"
    STRUCTCONVERT_CALL = "$SCHRODINGER/utilities/structconvert"
    FMP_STATS = "fmp_stats"
    FMP_STATS_CALL = "$SCHRODINGER/run -FROM scisol fmp_stats.py"
    STRUCTCAT_HELP = "-h"
    PROTEIN_INTERACTION_CALL = "$SCHRODINGER/run protein_interaction_analysis.py"
    PROTEIN_INTERACTION = "protein_interaction"
    STRUCTCAT_HELP_IDENTIFICATION_STRING = "<format> must be one of"
    # input; note that the format is directly appended (e.g. "-isd")
    STRUCTCAT_I = "-i"
    STRUCTCAT_O = (
        # output; note that the format is directly appended (e.g. "-omae")
        "-o"
    )
    STRUCTCAT_FORMAT_MAE = "mae"  # Maestro format
    STRUCTCAT_FORMAT_SD = "sd"  # MDL SDfile format
    STRUCTCAT_FORMAT_SDF = "sdf"
    STRUCTCAT_FORMAT_PDB = "pdb"  # PDB format
    STRUCTCAT_FORMAT_MOL2 = "mol2"  # sybyl (.mol2) format
    POSEVIEWER_FILE_KEY = "pv.maegz"
    PREPWIZARD = "prepwizard"
    PREPWIZARD_CALL = "$SCHRODINGER/utilities/prepwizard"
    MULTISIM_EXEC = "$SCHRODINGER/utilities/multisim"
    AWS_BINARY_LOC = "ssh <location> /opt/schrodinger/suite/installations/default/"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class PrimeEnum:

    PRIME_MMGBSA = "prime_mmgbsa"
    PRIME_HELP = "-h"
    PRIME_HELP_IDENTIFICATION_STRING = (
        "run $SCHRODINGER/prime_mmgbsa -h for a complete listing of all options."
    )
    PRIME_NJOBS = "-NJOBS"
    PRIME_WAIT = "-WAIT"

    # settings
    PRIME_OUTTYPE = "-out_type"
    PRIME_OUTTYPE_LIGAND = "LIGAND"

    # tags in output
    PRIME_MMGBSA_TOTAL_ENERGY = (
        "r_psp_MMGBSA_dG_Bind"  # total energy of binding: complex - receptor - ligand
    )
    PRIME_MMGBSA_TOTAL_ENERGY_NS = (
        # as above but without strain energy correction
        "r_psp_MMGBSA_dG_Bind(NS)"
    )

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class PantherEnum:

    PANTHER_PTYHON2 = "python2"
    PANTHER_ENTRYPOINT = "panther.py"
    PANTHER_CONFIG = "panther_config.in"
    PANTHER_OUTPUT_FILE = "neg_image.mol2"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class KallistoEnum:

    KALLISTO = "kallisto"  # Binary name.

    # OPTIONS
    SILENT = "--silent"
    SHIFT = "--shift"  # INTEGER argument
    HELP = "--help"  # Show this message and exit
    HELP_IDENTIFICATION_STRING = "Show this message"

    # COMMANDS
    ALP = "alp"  # Static atomic polarizabilities in Bohr^3.
    BONDS = "bonds"  # Get information about covalent bonding partner.
    CNS = "cns"  # Atomic coordination numbers.
    EEQ = "eeq"  # Electronegativity equilibration atomic partial charges.
    EXS = (
        "exs"  # Exchange a substrate within a transition metal complex with another...
    )
    LIG = "lig"  # Get all substructures (or ligands) that are bound to the center...
    PROX = "prox"  # Atomic proximity shells.
    RMS = "rms"  # Calculate the root mean squared deviation between two structures...
    SORT = "sort"  # Sort input geoemtry according to connectivity.
    STM = "stm"  # Calculate sterimol descriptors using kallisto van der Waals radii.
    VDW = "vdw"  # Charge-dependent atomic van der Waals radii in Bohr.

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class JazzyEnum:
    JAZZY = "jazzy"
    HELP = "--help"
    HELP_IDENTIFICATION_STRING = "Show this message"

    # "vec" command
    VEC = "vec"  # "generating properties" mode
    VEC_OPT = "--opt"  # optimization
    VEC_OPT_MMF94 = "MMF94"
    VEC_OPT_MMF94S = "MMF94s"
    VEC_OPT_UFF = "UFF"
    VEC_STRENGHT_ONLY = "--strength_only"

    # "vis" command
    VIS = "vis"  # "generating graphics" mode
    VIS_OPT = "--opt"
    VIS_OPT_MMF94 = "MMF94"
    VIS_OPT_MMF94S = "MMF94s"
    VIS_OPT_UFF = "UFF"
    VIS_FIG_SIZE = "--fig_size"  # Size of SVG image in pixels.  [default: 500, 500]
    VIS_SDC_THRESHOLD = (
        "--sdc_threshold"  # Treshold strength to depic Carbon donors. [default: 0.0]
    )
    VIS_SA_THRESHOLD = (
        "--sa_threshold"  # Treshold strength to depic acceptors. [default: 0.0]
    )
    VIS_BASE64 = "--base64"
    VIS_FLATTEN_MOLECULE = "--flatten_molecule"
    VIS_HIGHLIGHT_ATOMS = "--highlight_atoms"
    VIS_IGNORE_SDC = "--ignore_sdc"
    VIS_IGNORE_SDX = "--ignore_sdx"
    VIS_IGNORE_SA = "--ignore_sa"
    VIS_HELP = "--help"

    # result dictionary keys for "vec" command
    RESULT_SDC = "sdc"
    RESULT_SDX = "sdx"
    RESULT_SA = "sa"
    RESULT_DGA = "dga"
    RESULT_DGP = "dgp"
    RESULT_DGTOT = "dgtot"
    RESULT_STATUS = "__status"
    RESULT_STATUS_SUCCESS = "success"
    RESULT_SMILES = "smiles"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class ShaepEnum:

    SHAEP_EXECUTABLE = "shaep"
    OUTPUT_SIMILARITY = "similarity.txt"
    NEGATIVE_IMAGE_OUTPUT_FILE = "neg_image.mol2"
    CONFORMER_PATH = "conformer.sdf"
    TAG_SHAPE_SIMILARITY = "shape_similarity"
    TAG_ESP_SIMILARITY = "esp_similarity"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class GromacsEnum:

    # gmx programs
    PDB2GMX = "gmx pdb2gmx"
    EDITCONF = "gmx editconf"
    SOLVATE = "gmx solvate"
    GROMPP = "gmx grompp"
    GENION = "gmx genion"
    MDRUN = "gmx mdrun"
    MPI_MDRUN = "gmx_mpi mdrun"
    MAKE_NDX = "gmx make_ndx"
    GENRESTR = "gmx genrestr"
    TRJCAT = "gmx trjcat"
    TRJCONV = "gmx trjconv"
    CLUSTER = "gmx cluster"
    RMS = "gmx rms"
    ANTECHAMBER = "antechamber"
    PDB2GMX_FAIL_ID_STRING = "Required option was not provided"
    ACPYPE_BINARY = "acpype"
    MMPBSA = "gmx_MMPBSA"
    DO_DSSP = "gmx do_dssp"
    SELECT = "gmx select"
    CLUSTER_TS = "Rscript $MDPLOT/MDplot/inst/bash/MDplot_bash.R clusters_ts"
    PRIMARY_COMPONENTS = ["Protein", "DNA", "RNA"]
    # from residuetypes.dat
    AMBER_PARAMETRISED_COMPONENTS = [
        "ABU",
        "ACE",
        "AIB",
        "ALA",
        "ARG",
        "ARGN",
        "ASN",
        "ASN1",
        "ASP",
        "ASP1",
        "ASPH",
        "ASPP",
        "ASH",
        "CT3",
        "CYS",
        "CYS1",
        "CYS2",
        "CYSH",
        "DALA",
        "GLN",
        "GLU",
        "GLUH",
        "GLUP",
        "GLH",
        "GLY",
        "HIS",
        "HIS1",
        "HISA",
        "HISB",
        "HISH",
        "HISD",
        "HISE",
        "HISP",
        "HSD",
        "HSE",
        "HSP",
        "HYP",
        "ILE",
        "LEU",
        "LSN",
        "LYS",
        "LYSH",
        "MELEU",
        "MET",
        "MEVAL",
        "NAC",
        "NME",
        "NHE",
        "NH2",
        "PHE",
        "PHEH",
        "PHEU",
        "PHL",
        "PRO",
        "SER",
        "THR",
        "TRP",
        "TRPH",
        "TRPU",
        "TYR",
        "TYRH",
        "TYRU",
        "VAL",
        "PGLU",
        "HID",
        "HIE",
        "HIP",
        "LYP",
        "LYN",
        "CYN",
        "CYM",
        "CYX",
        "DAB",
        "ORN",
        "HYP",
        "NALA",
        "NGLY",
        "NSER",
        "NTHR",
        "NLEU",
        "NILE",
        "NVAL",
        "NASN",
        "NGLN",
        "NARG",
        "NHID",
        "NHIE",
        "NHIP",
        "NHISD",
        "NHISE",
        "NHISH",
        "NTRP",
        "NPHE",
        "NTYR",
        "NGLU",
        "NASP",
        "NLYS",
        "NORN",
        "NDAB",
        "NLYSN",
        "NPRO",
        "NHYP",
        "NCYS",
        "NCYS2",
        "NMET",
        "NASPH",
        "NGLUH",
        "CALA",
        "CGLY",
        "CSER",
        "CTHR",
        "CLEU",
        "CILE",
        "CVAL",
        "CASN",
        "CGLN",
        "CARG",
        "CHID",
        "CHIE",
        "CHIP",
        "CHISD",
        "CHISE",
        "CHISH",
        "CTRP",
        "CPHE",
        "CTYR",
        "CGLU",
        "CASP",
        "CLYS",
        "CORN",
        "CDAB",
        "CLYSN",
        "CPRO",
        "CHYP",
        "CCYS",
        "CCYS2",
        "CMET",
        "CASPH",
        "CGLUH",
        "DA",
        "DG",
        "DC",
        "DT",
        "DA5",
        "DG5",
        "DC5",
        "DT5",
        "DA3",
        "DG3",
        "DC3",
        "DT3",
        "DAN",
        "DGN",
        "DCN",
        "DTN",
        "A",
        "U",
        "C",
        "G",
        "RA",
        "RU",
        "RC",
        "RG",
        "RA5",
        "RT5",
        "RU5",
        "RC5",
        "RG5",
        "RA3",
        "RT3",
        "RU3",
        "RC3",
        "RG3",
        "RAN",
        "RTN",
        "RUN",
        "RCN",
        "RGN",
        "SOL",
        "WAT",
        "HOH",
        "OHH",
        "TIP",
        "T3P",
        "T4P",
        "T5P",
        "T3H",
        "K",
        "NA",
        "CA",
        "MG",
        "CL",
        "ZN",
        "CU1",
        "CU",
        "LI",
        "NA+",
        "RB",
        "CS",
        "F",
        "CL-",
        "BR",
        "I",
        "OH",
        "Cal",
        "IB+",
    ]
    IONS = ["ZN", "MG", "CU", "CA", "NA", "CL", "RB", "CS", "F", "BR", "I", "OH", "K"]
    LIG_ID = "lig_id.lig"
    LIG_EXT = "lig"
    ATOMS = ["HETATM", "ATOM"]
    ATOMS_DIRECTIVE = "[ atoms ]"
    BONDS = "[ bonds ]"
    ATOMTYPES = "[ atomtypes ]"
    MOLECULETYPES = "[ moleculetype ]"
    MOLECULES = "[ molecules ]"
    SOLVENTS = ["HOH ", "SOL", "WAT"]
    TERMINATIONS = ["ENDMDL", "END"]
    SYSTEM = "[ system ]"
    DEFAULTS = "[ defaults ]"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class PMXEnum:

    # $PMX programs (see respective steps for the help strings)
    ABFE = "$PMX abfe"
    ANALYSE = "$PMX analyse"
    ATOMMAPPING = "$PMX atomMapping"
    DOUBLEBOX = "$PMX doublebox"
    GENLIB = "$PMX genlib"
    GENTOP = "$PMX gentop"
    LIGANDHYBRID = "$PMX ligandHybrid"
    MUTATE = "$PMX mutate"

    # custom scripts
    BOX_WATER_IONS = "box_water_ions.py"
    PREPARE_SIMULATIONS = "prepare_simulations.py"
    PREPARE_TRANSITIONS = "prepare_transitions.py"
    RUN_ANALYSIS = "run_analysis.py"
    RUN_SIMULATIONS = "run_simulations.py"
    ASSEMBLE_SYSTEMS = "assemble_systems.py"

    ANALYSE_HELP = "-h"
    ANALYSE_HELP_SUCCESS_STRING = "Calculates free energies from fast"

    # standard file extensions
    PDB = "pdb"

    # file system standards
    HYBRID_STR_TOP = "hybridStrTop"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepPMXEnum:
    SIM_TYPE = "sim_type"
    BOXSHAPE = "boxshape"
    BOXD = "boxd"
    WATER = "water"
    CONC = "conc"
    PNAME = "pname"
    NNAME = "nname"

    RUN_TYPE = "run_type"
    SIM_TYPES = "sim_types"
    EXEC_MODE = "exec_mode"
    MDRUN_EXECUTABLE = "mdrun_executable"

    ABFE = "abfe"
    RBFE = "rbfe"

    PREV_STEP = "previous_step"
    STRICT = "strict"


class PMXAtomMappingEnum:

    HELP = "--help"  # show this help message and exit. <>
    I1 = "-i1"  # Input ligand structure 1. Default is "lig1.pdb" <1 pdb file>
    I2 = "-i2"  # Input ligand structure 2. Default is "lig2.pdb" <1 pdb file>
    O1 = "-o1"  # Output pairs: column1:mol1, column2:mol2. Default is "pairs1.dat" <1 dat file>
    O2 = "-o2"  # Output pairs: column1:mol2, column2:mol1. Default is "pairs2.dat" <1 dat file>
    OPDB1 = "-opdb1"  # Optional output: superimposed structure 1. <1 pdb file>
    OPDB2 = "-opdb2"  # Optional output: superimposed structure 2. <1 pdb file>
    OPDBM1 = "-opdbm1"  # Optional output: morphable atoms in structure 1 <2 pdb files>
    OPDBM2 = "-opdbm2"  # Optional output: morphable atoms in structure 2. <2 pdb files>
    # Optional output: score of the morph. Default is "out_score.dat" <1 dat file>
    SCORE = "-score"
    N1 = "-n1"  # Optional input: index of atoms to consider for mol1 <1 ndx file>
    N2 = "-n2"  # Optional input: index of atoms to consider for mol2 <1 ndx file>
    LOG = "-log"  # Output: log file. Default is "mapping.log" <1 log file>
    NO_ALIGNMENT = (
        # Should the alignment method be disabled (default enabled) <>
        "--no-alignment"
    )
    # Should the MCS method be disabled (default enabled) <>
    NO_MCS = "--no-mcs"
    # Should non-polar hydrogens be discarded from morphing into any other hydrogen (default True) <>
    NO_H2H = "--no-H2H"
    # Should polar hydrogens be morphed into polar hydrogens (default False) <>
    H2HPOLAR = "--H2Hpolar"
    H2HEAVY = (
        # Should hydrogen be morphed into a heavy atom (default False) <>
        "--H2Heavy"
    )
    # Should rings only be used in the MCS search and alignemnt (default False) <>
    RINGSONLY = "--RingsOnly"
    # Should the distance criterium be also applied in the MCS based search (default False) <>
    DMCS = "--dMCS"
    # Try swapping the molecule order which would be a cross-check and require double execution time (default False) <>
    SWAP = "--swap"
    NO_CHIRALITY = (
        # Perform chirality check for MCS mapping (default True) <>
        "--no-chirality"
    )
    # Distance (nm) between atoms to consider them morphable for alignment approach (default 0.05 nm). <1 numeric value>
    D = "--d"
    # Maximum time (s) for an MCS search (default 10 s). <1 numeric value>
    TIMEOUT = "--timeout"

    LIGAND_DIR = "input/ligands"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class PMXLigandHybridEnum:

    HELP = "--help"  # Show this help message and exit. <>
    I1 = "-i1"  # Input ligand structure 1. Default is "lig1.pdb" <1 pdb file>
    I2 = "-i2"  # Input ligand structure 2. Default is "lig2.pdb" <1 pdb file>
    ITP1 = "-itp1"  # Input ligand topology 1. Default is "lig1.itp" <1 itp file>
    ITP2 = "-itp2"  # Input ligand topology 2. Default is "lig2.itp" <1 itp file>
    PAIRS = "-pairs"  # Optional input: atom pair mapping. <1 dat file>
    N1 = "-n1"  # Optional input: index of atoms to consider for mol1. <1 ndx file>
    N2 = "-n2"  # Optional input: index of atoms to consider for mol2. <1 ndx file>
    OA = "-oA"  # Output: hybrid structure based on the ligand 1. Default is "mergedA.pdb" <1 pdb file>
    OB = "-oB"  # Output: hybrid structure based on the ligand 2. Default is "mergedB.pdb" <1 pdb file>
    OITP = "-oitp"  # Output: hybrid topology. Default is "merged.itp". <1 itp file>
    # Output: atomtypes for hybrid topology. Default is "ffmerged.itp" <1 itp file>
    OFFITP = "-offitp"
    LOG = "-log"  # Output: log file. Default is "hybrid.log" <1 log file>
    # Optional: if -pairs not provided, distance (nm) between atoms to consider them morphable
    # for alignment approach (default 0.05 nm). <1 numerical value>
    D = "--d"
    FIT = "--fit"  # Fit mol2 onto mol1, only works if pairs.dat is provided. <>
    SPLIT = "--split"  # Split the topology into separate transitions.
    SCDUMM = (
        "--scDUMm"  # Scale dummy masses using the counterpart atoms. <1 numeric value>
    )
    SCDUMA = "--scDUMa"  # Scale bonded dummy angle parameters. <1 numeric value>
    SCDUMD = "--scDUMd"  # Scale bonded dummy dihedral parameters. <1 numeric value>
    DEANG = "--deAng"  # Decouple angles composed of 1 dummy and 2 non-dummies.

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class FepPlusEnum:
    FEP_MAPPER = "$SCHRODINGER/run -FROM scisol fep_mapper.py"
    FEP_EXECUTOR = "$SCHRODINGER/fep_plus"
    FEP_ABSOLUTE_EXECUTOR = "$SCHRODINGER/fep_absolute_binding"
    FEP_HELP = "-h"
    JSC_LIST = 'ssh $SCHRODINGER_JOBSERVER "export SCHRODINGER=/opt/schrodinger/suite/installations/default && /opt/schrodinger/suite/installations/default/jsc list"'
    JSC_TAIL_FILE = 'ssh $SCHRODINGER_JOBSERVER "export SCHRODINGER=/opt/schrodinger/suite/installations/default && /opt/schrodinger/suite/installations/default/jsc tail-file'
    DICT = "dict"
    PATH = "path"
    FEP_MAPPER_HELP_SUCCESS_STRING = "If given, the match will be allowed"


class PdbFixerEnum:
    FIXER = "pdbfixer"


class DSSPEnum:
    MKDSSP = "mkdssp"


class SlurmEnum:
    SBATCH = "sbatch"
    STATE = "State"
    COMPLETED = "COMPLETED"
    RUNNING = "RUNNING"
    PENDING = "PENDING"
    CANCELLED = "CANCELLED"
    NODE_FAIL = "NODE_FAIL"
    FAILED = "FAILED"



from icolos.utils.execute_external.execute import Executor
from icolos.core.workflow_steps.calculation.base import StepCalculationBase
from icolos.core.workflow_steps.step import _LE
from icolos.utils.enums.step_enums import StepShaepEnum
from icolos.utils.enums.program_parameters import PantherEnum, ShaepEnum
from icolos.core.containers.compound import Conformer
import tempfile
from pydantic import BaseModel
import os

_SSE = StepShaepEnum()
_SE = ShaepEnum()
_PE = PantherEnum()


class StepShaep(StepCalculationBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=Executor)

    def _prepare_tmp_input_dir(self):
        tmp_dir = tempfile.mkdtemp()
        return tmp_dir

    def _execute_backend(self, conf_path: str, tmp_dir: str, ni_path: str):
        arguments = [
            os.path.join(tmp_dir, ni_path),
            conf_path,
            os.path.join(tmp_dir, _SE.OUTPUT_SIMILARITY),
        ]
        self._backend_executor.execute(
            command=_SE.SHAEP_EXECUTABLE, arguments=arguments, check=True
        )

    def _parse_output(self, tmp_dir: str, conformer: Conformer):
        with open(os.path.join(tmp_dir, _SE.OUTPUT_SIMILARITY), "r") as f:
            # TODO: add support for multiple input structures; ignore the names (all will be in one line), but from
            #       position 8 (index 7 in python) onwards, the shape and esp similarities are reported in the same
            #       order as the input, i.e. <7 other values> mol1_shape mol1_esp mol2_shape ...
            parts = f.readlines()[1].split("\t")
            conformer.get_molecule().SetProp(_SE.TAG_SHAPE_SIMILARITY, str(parts[7]))
            conformer.get_molecule().SetProp(_SE.TAG_ESP_SIMILARITY, str(parts[8]))

    def execute(self):
        number_rescored = 0
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if len(enumeration.get_conformers()) == 0:
                    self._logger.log(
                        f"Found no conformers for enumeration {enumeration} for compound {compound}.",
                        _LE.WARNING,
                    )
                    # we can still execute shaep at the enumeration level, if the compounds are correcty annotated they should be written out ok.  Will be slower though
                    # easiest for now is to add the enumeration mol object as a single conformer and run that through shaep
                    mol = enumeration.get_molecule()
                    conf = Conformer(conformer=mol)
                    enumeration.add_conformer(conf)

                # TODO: ShaEP allow batch execution for any number of compounds (parsing gets more difficult though)
                #       Implement that to avoid overhead from file system issues
                # TODO: Refactor and add comments
                for conformer in enumeration.get_conformers():
                    tmp_dir = self._prepare_tmp_input_dir()
                    conf_path = os.path.join(tmp_dir, _SE.CONFORMER_PATH)
                    ni_file = self.data.generic.get_files_by_extension("mol2")[0]
                    ni_file.write(tmp_dir)
                    conformer.write(conf_path)
                    self._execute_backend(conf_path, tmp_dir, ni_file.get_file_name())
                    self._parse_output(tmp_dir, conformer)
                    self._logger.log(
                        f"Finished shaep execution for conformer {enumeration.get_index_string()}.",
                        _LE.DEBUG,
                    )
                    number_rescored += 1
                    self._remove_temporary(tmp_dir)
        self._logger.log(f"Executed ShaEP for {number_rescored} conformers.", _LE.INFO)



import tempfile
from typing import Callable, List
from icolos.core.containers.generic import GenericData
from icolos.core.containers.gmx_state import GromacsState
from icolos.utils.enums.execution_enums import ExecutionPlatformEnum
from icolos.utils.enums.step_enums import StepGromacsEnum
from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
import os

from icolos.utils.execute_external.gromacs import GromacsExecutor
from icolos.utils.general.parallelization import Parallelizer, SubtaskContainer

_GE = GromacsEnum()
_SGE = StepGromacsEnum()
_ERE = ExecutionPlatformEnum


class StepGMXMDrun(StepGromacsBase, BaseModel):
    """
    Launch gmx mdrun
    """

    topol: GromacsState = None

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()

    def _get_log_file(self, tmp_dir):
        """
        Find and parse the log file
        """
        log_file = [f for f in os.listdir(tmp_dir) if f.endswith(".log")]
        assert len(log_file) == 1
        with open(os.path.join(tmp_dir, log_file[0]), "r") as f:
            data = f.readlines()
        return data

    def _tail_log_file(self, tmp_dir):
        """
        Log the last 50 lines of the log file to capture performance metrics from the run

        """
        log_file = self._get_log_file(tmp_dir)

        for line in log_file[-50:]:
            self._logger_blank.log(line, _LE.INFO)

    def execute_mdrun(self, path: str, index: int):
        """
        Make a single call to mdrun
        """
        flag_dict = (
            {
                "-s": _SGE.STD_TPR,
                "-c": _SGE.STD_STRUCTURE,
                "-e": _SGE.STD_EDR,
                "-cpo": _SGE.STD_CPT,
                "-x": _SGE.STD_XTC,
            }
            if not self.data.generic.get_files_by_extension("cpt")
            else {"-cpi", os.path.join(path, "state.cpt")}
        )

        arguments = self._parse_arguments(flag_dict)
        self._backend_executor.execute(
            command=_GE.MDRUN, arguments=arguments, location=path, check=True
        )

    def execute_parallel_simulations(self, work_dirs, run_func: Callable):
        # attach the index of the workdir
        work_dirs = [(idx, wkdir) for idx, wkdir in enumerate(work_dirs)]
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(work_dirs)
        parallelizer = Parallelizer(func=run_func)
        n = 1
        while self._subtask_container.done() is False:
            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())
            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]
            paths, indices = self.prepare_jobs(next_batch)
            parallelizer.execute_parallel(path=paths, index=indices)
            n += 1

    def prepare_jobs(self, batch) -> List[tuple]:
        paths, indices = [], []
        for task in batch:
            for element in task:
                # tuple of (idx, dirpath)
                paths.append(element.data[1])
                indices.append(element.data[0])
        return paths, indices

    def run_single_tpr(self, tmp_dir: str):
        """
        Normal gmx mdrun call, if multiple structures are loaded into the topology, run them in parallel according to the parallelizer settings
        """
        # if we have multiple structures, run the simulations externally, in parallel
        work_dirs = [tempfile.mkdtemp(dir=tmp_dir) for _ in range(len(self.topol.tprs))]

        # prepare tmpdirs with tpr files
        for path, tpr in zip(work_dirs, self.topol.tprs.values()):
            tpr.write(path)

        # if > 1, instantiate a parallelizer, load the paths in and execute in parallel, user should be using the slurm/SGE interface to request extern resources
        if len(work_dirs) > 1:
            self.execute_parallel_simulations(work_dirs, run_func=self.execute_mdrun)
        else:
            tmp_dir = work_dirs[0]
            self.execute_mdrun(tmp_dir, index=0)

        # now parse the outputs
        for index, path in enumerate(work_dirs):
            # set a structure other than confout.gro e.g. if a pdb output has been set
            struct = (
                self.settings.arguments.parameters["-c"]
                if "-c" in self.settings.arguments.parameters.keys()
                else _SGE.STD_STRUCTURE
            )
            self.topol.set_structure(path, file=struct, index=index)
            self.topol.set_trajectory(path, index=index)
            self.topol.set_log(path, index=index)
            self.topol.set_edr(path, index=index)
            try:
                self.topol.set_cpt(path, index=index)
            except FileNotFoundError:
                self._logger.log("No checkpoint file generated", _LE.DEBUG)

    def run_multidir_sim(self, tmp_dir: str):
        """
        Runs a multidir simulation, allowing for replex simulations.  Several conditions are required for this running mode
        1) the previous step in the workflow should have been an iterator to produce n tpr files.  This must have been run with single_dir mode ON and remove_temprorary_files OFF, so we can extract files from those workflows' tmpdirs
        """
        if not self.execution.platform == _ERE.SLURM:
            self._logger.log(
                "WARNING: Running HREX simulation without external resources! Normally this should be run as a separate batch job",
                _LE.WARNING,
            )

        # extract the tprs from the topol object, write to separate tmpdirs
        work_dirs = [tempfile.mkdtemp(dir=tmp_dir) for _ in range(len(self.topol.tprs))]
        self._logger.log(
            f"Initiating gmx multidir run in directories {', '.join(work_dirs)}",
            _LE.DEBUG,
        )
        for path, tpr in zip(work_dirs, self.topol.tprs.values()):
            tpr.write(path)

        # note, this must be a multiple of the number of simulations
        tasks = self.execution.resources.tasks
        # map the PP and PME tasks to the GPUs

        command = f"mpirun -np {tasks} gmx_mpi mdrun -multidir {' '.join(work_dirs)}"
        arguments = self._parse_arguments(flag_dict={"-x": _SGE.STD_XTC})
        self._backend_executor.execute(
            command=command, arguments=arguments, location=tmp_dir, check=True
        )
        # udpate the structures to the new coordinates
        for i, work_dir in enumerate(work_dirs):
            self.topol.set_structure(work_dir, index=i)
            self.topol.set_trajectory(work_dir, index=i)
            self.topol.set_tpr(work_dir, index=i)
            self.topol.set_log(work_dir, index=i)
            self.topol.set_edr(path, index=i)
            try:
                self.topol.set_cpt(path, index=i)
            except FileNotFoundError:
                self._logger.log("No checkpoint file generated", _LE.DEBUG)

    def execute(self):

        tmp_dir = self._make_tmpdir()
        self.topol = self.get_topol()
        if self.data.generic.get_files_by_extension("cpt"):
            # a cpt file has been passed, simply restart
            print(self.data.generic.get_file_names_by_extension("gpt"))
            self._run_checkpoint_files(self.data.generic.get_files_by_extension("cpt"))

        self.execution.parallelization.max_length_sublists = 1
        # pickle the topol to the mdrun dir, if something goes wrong/the job dies, the workflow can be picked up where we left off by unpickling the topology object
        self.pickle_topol(self.topol, tmp_dir)
        multidir = self._get_additional_setting(_SGE.MULTIDIR, default=False)
        if multidir:
            self.run_multidir_sim(tmp_dir)
        else:
            self.run_single_tpr(tmp_dir)
        self._remove_temporary(tmp_dir)



from icolos.utils.enums.step_enums import StepGromacsEnum, StepPrepwizEnum
from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum
from icolos.core.workflow_steps.schrodinger.base import StepSchrodingerBase
from icolos.utils.execute_external.schrodinger import SchrodingerExecutor
from icolos.core.containers.generic import GenericData
from pydantic import BaseModel
from copy import deepcopy
import os

_SEE = SchrodingerExecutablesEnum()
_SGE = StepGromacsEnum()
_SPE = StepPrepwizEnum()


class StepPrepwizard(StepSchrodingerBase, BaseModel):
    """
    Interface to Schrodinger's PrepWizard program for protein prep
    """

    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(executor=SchrodingerExecutor)
        self._check_backend_availability()

    def _parse_args(self):
        parameters = deepcopy(self.settings.arguments.parameters)
        arguments = []
        if len(self.settings.arguments.flags) > 0:
            for flag in self.settings.arguments.flags:
                arguments.append(str(flag))
        if parameters:
            for key in parameters.keys():
                arguments.append(key)
                if parameters[key] is not None and parameters[key] != "":
                    arguments.append(str(parameters[key]))
        input_file = self.data.generic.get_file_names_by_extension("pdb")[0]
        output_file = input_file  # write to the same file name to keep things tidy
        arguments.append(input_file)
        arguments.append(output_file)
        return arguments

    def _parse_output(self, tmp_dir: str):
        output_pdb = os.path.join(
            tmp_dir, self.data.generic.get_file_names_by_extension("pdb")[0]
        )
        with open(output_pdb, "r") as f:
            data = f.read()
        self.data.generic.clear_file_dict()
        output_file = GenericData(file_name=_SGE.COMPLEX_PDB, file_data=data)
        self.data.generic.add_file(output_file)

    def _remove_ligand(self, tmp_dir):
        remove_res = self.settings.additional[_SPE.REMOVE_RES]
        pdb_file = self.data.generic.get_argument_by_extension("pdb")
        cleaned_pdb_lines = []
        # handle ligand removal mode: strip ligands, leave cofactors
        if remove_res != _SPE.LIGANDS and not isinstance(remove_res, list):
            remove_res = list(remove_res)

        with open(os.path.join(tmp_dir, pdb_file), "r") as f:
            if remove_res == _SPE.LIGANDS:
                # automatically remove ligands, keep cofactors that are specified in the enum.
                for line in f.readlines():
                    if (
                        line is not None
                        and len(line.split()) > 3
                        and (
                            line.split()[0] == "ATOM"
                            or any(l in line for l in _SPE.COFACTOR_IDS)
                        )
                    ):
                        cleaned_pdb_lines.append(line)
            else:
                for line in f.readlines():
                    if not any(l in line for l in remove_res):
                        cleaned_pdb_lines.append(line)

        with open(os.path.join(tmp_dir, pdb_file), "w") as f:
            f.writelines(cleaned_pdb_lines)

    def execute(self):
        tmp_dir = self._make_tmpdir()
        args = self._parse_args()
        self.data.generic.write_out_all_files(tmp_dir)
        if (
            _SPE.REMOVE_RES in self.settings.additional.keys()
            and self.settings.additional[_SPE.REMOVE_RES] is not None
        ):
            self._remove_ligand(tmp_dir)
        self._backend_executor.execute(
            command=_SEE.PREPWIZARD, arguments=args, check=True, location=tmp_dir
        )

        self._parse_output(tmp_dir)



from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum
from icolos.utils.execute_external.execute import ExecutorBase

_SEE = SchrodingerExecutablesEnum()


class SchrodingerExecutor(ExecutorBase):
    """For the execution of Schrodinger's support entry points"""

    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        # check, whether a proper executable is provided; update the calls to "$SCHRODINGER/XYZ"
        if command == _SEE.STRUCTCAT:
            command = _SEE.STRUCTCAT_CALL
        elif command == _SEE.SDCONVERT:
            command = _SEE.SDCONVERT_CALL
        elif command == _SEE.STRUCT_SPLIT:
            command = _SEE.STRUCT_SPLIT_CALL
        elif command == _SEE.STRUCTCONVERT:
            command = _SEE.STRUCTCONVERT_CALL
        elif command == _SEE.FMP_STATS:
            command = _SEE.FMP_STATS_CALL
        elif command == _SEE.PREPWIZARD:
            command = _SEE.PREPWIZARD_CALL
        elif command == _SEE.MULTISIM_EXEC:
            command = _SEE.MULTISIM_EXEC
        elif command == _SEE.PROTEIN_INTERACTION:
            command = _SEE.PROTEIN_INTERACTION_CALL
        else:
            raise ValueError(
                "Parameter command must be an dictionary of the internal Schrodinger entry point list."
            )

        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=location,
            pipe_input=pipe_input,
        )

    def is_available(self):
        try:
            result = self.execute(
                command=_SEE.STRUCTCAT, arguments=[_SEE.STRUCTCAT_HELP], check=True
            )

            if _SEE.STRUCTCAT_HELP_IDENTIFICATION_STRING in result.stdout:
                return True
            return False
        except Exception as e:

            print(e)
            return False



import json
import os
import numpy as np
import pandas as pd
from collections import OrderedDict
from copy import deepcopy
from typing import Tuple, List

from pydantic import BaseModel

from icolos.core.containers.compound import Conformer
from icolos.core.containers.generic import GenericData
from icolos.utils.enums.program_parameters import ModelBuilderEnum
from icolos.utils.enums.step_enums import StepModelBuilderEnum
from icolos.core.workflow_steps.io.base import StepIOBase
from icolos.core.workflow_steps.step import _LE, StepSettingsParameters
from icolos.utils.enums.write_out_enums import WriteOutEnum
from icolos.utils.execute_external.execute import Executor

_SMBE = StepModelBuilderEnum()
_SME = ModelBuilderEnum()
_WE = WriteOutEnum()


class StepModelBuilder(StepIOBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor
        self._initialize_backend(executor=Executor)

    def _generate_temporary_input_output_files(
        self, tmp_dir: str
    ) -> Tuple[str, str, str, str, str]:
        tmp_input_config_json = os.path.join(tmp_dir, _SMBE.TMP_INPUT_CONFIG)
        tmp_input_data_csv = os.path.join(tmp_dir, _SMBE.TMP_INPUT_DATA)
        tmp_output_best_model_pkl = os.path.join(tmp_dir, _SMBE.TMP_OUTPUT_BEST_MODEL)
        tmp_output_best_parameters_json = os.path.join(
            tmp_dir, _SMBE.TMP_OUTPUT_BEST_PARAMETERS
        )
        tmp_output_production_pkl = os.path.join(
            tmp_dir, _SMBE.TMP_OUTPUT_PRODUCTION_MODEL
        )
        return (
            tmp_input_config_json,
            tmp_input_data_csv,
            tmp_output_best_model_pkl,
            tmp_output_best_parameters_json,
            tmp_output_production_pkl,
        )

    def _update_data_block(
        self, conf: dict, tmp_input_data_csv: str, settings: StepSettingsParameters
    ) -> dict:
        # the user can specify additional things for the "data" block of the configuration
        # in the "additional" field; the input CSV file needs to be overwritten in every case, though
        specified_data_block = settings.additional.get(_SMBE.DATA, {})
        for key in specified_data_block.keys():
            conf[_SMBE.DATA][key] = specified_data_block[key]
        conf[_SMBE.DATA][_SMBE.DATA_TRAININGSET_FILE] = tmp_input_data_csv
        if _SMBE.DATA_TESTSET_FILE in conf[_SMBE.DATA].keys():
            conf[_SMBE.DATA].pop(_SMBE.DATA_TESTSET_FILE, None)
            self._logger.log(
                f"Removed test set specification, not supported yet.", _LE.WARNING
            )
        return conf

    def _write_OptunaAZ_configuration(
        self,
        tmp_input_config_json: str,
        tmp_input_data_csv: str,
        settings: StepSettingsParameters,
    ):
        config_path = settings.arguments.parameters[_SME.CONFIG]
        with open(config_path, "r") as file:
            optunaaz_conf = file.read().replace("\r", "").replace("\n", "")
            optunaaz_conf = json.loads(optunaaz_conf)
        optunaaz_conf = self._update_data_block(
            optunaaz_conf, tmp_input_data_csv, settings
        )
        with open(tmp_input_config_json, "w") as file:
            json.dump(optunaaz_conf, fp=file, indent=4)
        self._logger.log(
            f"Wrote updated OptunaAZ configuration file to {tmp_input_config_json}.",
            _LE.DEBUG,
        )

    def _write_input_csv(
        self,
        conformers: List[Conformer],
        tmp_input_data_csv: str,
        settings: StepSettingsParameters,
    ):
        def _get_tag(conformer: Conformer, tag: str) -> str:
            try:
                value = conformer.get_molecule().GetProp(tag).strip()
            except KeyError:
                value = np.nan
            return value

        smiles_column = settings.additional[_SMBE.DATA][_SMBE.DATA_INPUT_COLUMN]
        response_column = settings.additional[_SMBE.DATA][_SMBE.DATA_RESPONSE_COLUMN]

        # initialize the dictionary
        dict_result = OrderedDict()
        dict_result[_WE.RDKIT_NAME] = ["" for _ in range(len(conformers))]
        dict_result[smiles_column] = ["" for _ in range(len(conformers))]
        dict_result[response_column] = ["" for _ in range(len(conformers))]

        # populate the dictionary with the values
        for irow in range(len(conformers)):
            conf = conformers[irow]
            dict_result[_WE.RDKIT_NAME][irow] = conf.get_index_string()
            dict_result[smiles_column][irow] = _get_tag(conf, smiles_column)
            dict_result[response_column][irow] = _get_tag(conf, response_column)

        # do the writeout (after sanitation)
        df_result = pd.DataFrame.from_dict(dict_result)
        df_result.to_csv(
            path_or_buf=tmp_input_data_csv,
            sep=",",
            na_rep="",
            header=True,
            index=False,
            mode="w",
            quoting=None,
        )

    def _get_arguments(
        self,
        tmp_input_config_json: str,
        tmp_output_best_model_pkl: str,
        tmp_output_best_parameters_json: str,
        tmp_output_production_pkl: str,
    ) -> List[str]:
        arguments = [
            _SME.CONFIG,
            tmp_input_config_json,
            _SME.MERGED_MODEL_OUTPATH,
            tmp_output_production_pkl,
            _SME.BEST_MODEL_OUTPATH,
            tmp_output_best_model_pkl,
            _SME.BEST_BUILDCONFIG_OUTPATH,
            tmp_output_best_parameters_json,
        ]
        return arguments

    def _parse_output(
        self,
        tmp_input_config_json: str,
        tmp_input_data_csv: str,
        tmp_output_best_parameters_json: str,
        tmp_output_production_pkl: str,
    ):
        # loading the final model is crucial (and the end-artifact for this step)
        try:
            with open(tmp_output_production_pkl, "rb") as f:
                data = f.read()
                self.data.generic.add_file(
                    GenericData(
                        file_name=_SMBE.TMP_OUTPUT_PRODUCTION_MODEL, file_data=data
                    )
                )
        except FileNotFoundError as e:
            self._logger.log(
                f"Could not load production model from path {tmp_output_production_pkl}.",
                _LE.ERROR,
            )
            raise e

        # loading the JSON with the best hyper-parameter configuration
        try:
            with open(tmp_output_best_parameters_json, "r") as f:
                data = f.read().replace("\r", "").replace("\n", "")
                data = json.loads(data)
                self.data.generic.add_file(
                    GenericData(
                        file_name=_SMBE.TMP_OUTPUT_BEST_PARAMETERS, file_data=data
                    )
                )
        except FileNotFoundError as e:
            self._logger.log(
                f"Could not load best hyper-parameter configuration from path {tmp_output_best_parameters_json}.",
                _LE.WARNING,
            )

        # loading the input JSON for OptunaAZ
        try:
            with open(tmp_input_config_json, "r") as f:
                data = f.read()
                self.data.generic.add_file(
                    GenericData(file_name=_SMBE.TMP_INPUT_CONFIG, file_data=data)
                )
        except FileNotFoundError as e:
            self._logger.log(
                f"Could not load input CSV file from path {tmp_input_config_json}.",
                _LE.WARNING,
            )

        # loading the input CSV
        try:
            with open(tmp_input_config_json, "r") as f:
                data = f.read()
                self.data.generic.add_file(
                    GenericData(file_name=_SMBE.TMP_INPUT_DATA, file_data=data)
                )
        except FileNotFoundError as e:
            self._logger.log(
                f"Could not load input CSV file from path {tmp_input_config_json}.",
                _LE.WARNING,
            )

    def execute(self):
        # make a copy of the settings to avoid side-effects with the dictionaries
        settings = deepcopy(self.settings)

        # generate temporary files
        tmp_dir = self._move_to_temp_dir()
        (
            tmp_input_config_json,
            tmp_input_data_csv,
            tmp_output_best_model_pkl,
            tmp_output_best_parameters_json,
            tmp_output_production_pkl,
        ) = self._generate_temporary_input_output_files(tmp_dir)

        # write OptunaAZ configuration to file
        self._write_OptunaAZ_configuration(
            tmp_input_config_json=tmp_input_config_json,
            tmp_input_data_csv=tmp_input_data_csv,
            settings=settings,
        )

        # unroll all conformers
        all_conformers = []
        for compound in self.get_compounds():
            for enumeration in compound:
                all_conformers = all_conformers + enumeration.get_conformers()

        # write input CSV, derived from the conformers
        self._write_input_csv(
            conformers=all_conformers,
            tmp_input_data_csv=tmp_input_data_csv,
            settings=settings,
        )

        # execute OptunaAZ
        self._backend_executor.execute(
            command=_SME.OPTBUILD_ENTRY_POINT,
            arguments=self._get_arguments(
                tmp_input_config_json=tmp_input_config_json,
                tmp_output_best_model_pkl=tmp_output_best_model_pkl,
                tmp_output_best_parameters_json=tmp_output_best_parameters_json,
                tmp_output_production_pkl=tmp_output_production_pkl,
            ),
            check=False,
        )

        # parse the output
        self._parse_output(
            tmp_input_config_json=tmp_input_config_json,
            tmp_input_data_csv=tmp_input_data_csv,
            tmp_output_best_parameters_json=tmp_output_best_parameters_json,
            tmp_output_production_pkl=tmp_output_production_pkl,
        )

        # clean-up
        self._restore_working_dir()
        self._remove_temporary(tmp_dir)



import os
from typing import List

from pydantic import BaseModel
from rdkit import Chem
from copy import deepcopy

from icolos.utils.execute_external.crest import CrestExecutor

from icolos.utils.general.molecules import get_charge_for_molecule

from icolos.core.containers.compound import Enumeration, Conformer

from icolos.utils.enums.program_parameters import CrestEnum, CrestOutputEnum
from icolos.core.workflow_steps.step import _LE, _CTE
from icolos.core.workflow_steps.confgen.base import StepConfgenBase

_EE = CrestEnum()
_COE = CrestOutputEnum()


class StepCREST(StepConfgenBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=CrestExecutor)
        self._check_backend_availability()

    def _get_energies_from_XYZ(self, path) -> list:
        energies = []
        with open(path, "r") as f:
            lines = f.readlines()
            for line in lines:
                if line.startswith(_COE.PREFIX_ENERGIES_XYZ):
                    energies.append(line.lstrip().rstrip())
        return energies

    def _parse_CREST_result(
        self, dir_path: str, enumeration: Enumeration
    ) -> List[Conformer]:
        """Function to parse the result from CREST."""
        # CREST will output a variety of files to "dir_path"
        conformers_sdf = os.path.join(dir_path, _COE.CREST_CONFORMERS_SDF)
        conformers_xyz = os.path.join(dir_path, _COE.CREST_CONFORMERS_XYZ)

        # as the energies are lost in the SDF output, we will add them as a tag
        energies = self._get_energies_from_XYZ(conformers_xyz)
        charge = str(
            get_charge_for_molecule(enumeration.get_molecule(), add_as_tag=False)
        )
        mol_supplier = Chem.SDMolSupplier(conformers_sdf, removeHs=False)
        result = []
        for mol_id, mol in enumerate(mol_supplier):
            mol.SetProp(_CTE.CONFORMER_ENERGY_TAG, energies[mol_id])
            mol.SetProp(_CTE.FORMAL_CHARGE_TAG, charge)
            result.append(Conformer(conformer=mol))
        return result

    def _set_formal_charge(self, parameters: dict, molecule: Chem.Mol) -> dict:
        charge = get_charge_for_molecule(molecule, add_as_tag=False)
        parameters[_EE.CREST_CHRG] = charge
        self._logger.log(f"Set charge for molecule to {charge}.", _LE.DEBUG)
        return parameters

    def _set_number_cores(self, parameters: dict) -> dict:
        """Function for parallelization of task, setting the number of cores to be used."""
        parameters[_EE.CREST_T] = int(self.execution.parallelization.jobs)
        return parameters

    def _prepare_settings(self, tmp_dir: str, enumeration: Enumeration) -> list:
        # first position is the input (SDF) file; the internal input at this stage is a molecule
        # -> write it to a temporary SDF file (undocumented input functionality) and add the path
        settings = [self._prepare_temp_input(tmp_dir, enumeration.get_molecule())]

        # add flags
        for flag in self.settings.arguments.flags:
            settings.append(flag)

        # add parameters
        parameters = deepcopy(self.settings.arguments.parameters)

        # update / over-write fields that need a specific value or are defined elsewhere
        parameters = self._set_number_cores(parameters)
        parameters = self._set_formal_charge(parameters, enumeration.get_molecule())

        # flatten the dictionary into a list for command-line execution
        for key in parameters.keys():
            settings.append(key)
            settings.append(parameters[key])
        return settings

    def execute(self):
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if not self._input_object_valid(enumeration):
                    continue

                # set up
                tmp_dir = self._move_to_temp_dir()

                # the call to CREST starts with the path to the input file, followed by arguments and flags
                settings = self._prepare_settings(tmp_dir, enumeration=enumeration)

                self._logger.log(
                    f"Executing CREST backend in folder {tmp_dir}.", _LE.DEBUG
                )
                self._backend_executor.execute(
                    command=_EE.CREST, arguments=settings, check=False
                )
                self._restore_working_dir()

                conformers = self._parse_CREST_result(tmp_dir, enumeration=enumeration)
                enumeration.clear_conformers()
                enumeration.add_conformers(conformers=conformers, auto_update=True)
                self._logger.log(
                    f"Executed CREST and obtained {len(conformers)} for enumeration {enumeration.get_index_string()}",
                    _LE.INFO,
                )

                self._remove_temporary(tmp_dir)



from typing import List
from pydantic import BaseModel

from icolos.core.containers.compound import Conformer, unroll_conformers
from icolos.utils.enums.step_enums import StepRMSDEnum, StepDataManipulationEnum
from icolos.core.workflow_steps.step import _LE
from icolos.core.workflow_steps.calculation.base import StepCalculationBase

_SR = StepRMSDEnum()
_SDM = StepDataManipulationEnum()


class StepRMSD(StepCalculationBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # extend parameters
        if _SR.METHOD not in self.settings.additional.keys():
            self.settings.additional[_SR.METHOD] = _SR.METHOD_ALIGNMOL

    def _calculate_RMSD(self, conformers: List[Conformer]):
        for conf in conformers:
            rmsd_matrix = self._calculate_rms_matrix(
                conformers=[conf] + conf.get_extra_data()[_SDM.KEY_MATCHED],
                rms_method=self._get_rms_method(),
            )

            # use the specified tag name if it is the first value and append an index in case there are more
            for idx, col in enumerate(rmsd_matrix.columns[1:]):
                combined_tag = "".join([_SR.RMSD_TAG, "" if idx == 0 else str(idx)])
                rmsd_value = rmsd_matrix.iloc[[0]][col][0]
                conf.get_molecule().SetProp(combined_tag, str(rmsd_value))
                conf.get_extra_data()[_SDM.KEY_MATCHED][idx].get_molecule().SetProp(
                    combined_tag, str(rmsd_value)
                )

    def execute(self):
        # this assumes that the conformers that are to be matched for the calculation of the RMSD matrix, are attached
        # as a list in a generic data field with a specified key
        conformers = unroll_conformers(compounds=self.get_compounds())
        self._calculate_RMSD(conformers=conformers)
        self._logger.log(
            f"Annotated {len(conformers)} conformers with RMSD values (tag: {_SR.RMSD_TAG}).",
            _LE.INFO,
        )

        # TODO: add a nice pandas DF with the RMSD values to a generic data field
